# Copyright (c) 2016 Santosh Philip
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================
"""useful functions that should in the class IDF. 
write them here first as a development step.
If they are really useful and don't muddy waters include them in IDF
and remove them here"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import itertools
from io import StringIO
from past.builtins import basestring  # pip install future

from eppy.modeleditor import IDF
from eppy.bunch_subclass import BadEPFieldError


def idfobjectkeys(idf):
    """returns the object keys in the order they were in the IDD file
    it is an ordered list of idf.idfobjects.keys()
    keys of a dict are unordered, so idf.idfobjects.keys() will not work for this purpose"""
    return idf.model.dtls


def getanymentions(idf, anidfobject):
    """Find out if idjobject is mentioned an any object anywhere"""
    name = anidfobject.obj[1]
    foundobjs = []
    keys = idfobjectkeys(idf)
    idfkeyobjects = [idf.idfobjects[key.upper()] for key in keys]
    for idfobjects in idfkeyobjects:
        for idfobject in idfobjects:
            if name.upper() in [
                item.upper() for item in idfobject.obj if isinstance(item, basestring)
            ]:
                foundobjs.append(idfobject)
    return foundobjs


def getobject_use_prevfield(idf, idfobject, fieldname):
    """field=object_name, prev_field=object_type. Return the object"""
    if not fieldname.endswith("Name"):
        return None
    # test if prevfieldname ends with "Object_Type"
    fdnames = idfobject.fieldnames
    ifieldname = fdnames.index(fieldname)
    prevfdname = fdnames[ifieldname - 1]
    if not prevfdname.endswith("Object_Type"):
        return None
    objkey = idfobject[prevfdname].upper()
    objname = idfobject[fieldname]
    try:
        foundobj = idf.getobject(objkey, objname)
    except KeyError as e:
        return None
    return foundobj


def getidfkeyswithnodes():
    """return a list of keys of idfobjects that hve 'None Name' fields"""
    idf = IDF(StringIO(""))
    keys = idfobjectkeys(idf)
    keysfieldnames = ((key, idf.newidfobject(key.upper()).fieldnames) for key in keys)
    keysnodefdnames = (
        (key, (name for name in fdnames if (name.endswith("Node_Name"))))
        for key, fdnames in keysfieldnames
    )
    nodekeys = [key for key, fdnames in keysnodefdnames if list(fdnames)]
    return nodekeys


def getobjectswithnode(idf, nodekeys, nodename):
    """return all objects that mention this node name"""
    keys = nodekeys
    # TODO getidfkeyswithnodes needs to be done only once. take out of here
    listofidfobjects = (
        idf.idfobjects[key.upper()] for key in keys if idf.idfobjects[key.upper()]
    )
    idfobjects = [idfobj for idfobjs in listofidfobjects for idfobj in idfobjs]
    objwithnodes = []
    for obj in idfobjects:
        values = obj.fieldvalues
        fdnames = obj.fieldnames
        for value, fdname in zip(values, fdnames):
            if fdname.endswith("Node_Name"):
                if value == nodename:
                    objwithnodes.append(obj)
                    break
    return objwithnodes


def name2idfobject(idf, groupnamess=None, objkeys=None, **kwargs):
    """return the object, if the Name or some other field is known.
    send field in ``**kwargs`` as Name='a name', Roughness='smooth'
    Returns the first find (field search is unordered)
    objkeys -> if objkeys=['ZONE', 'Material'], search only those
    groupnames -> not yet coded"""
    # TODO : this is a very slow search. revist to speed it up.
    if not objkeys:
        objkeys = idfobjectkeys(idf)
    for objkey in objkeys:
        idfobjs = idf.idfobjects[objkey.upper()]
        for idfobj in idfobjs:
            for key, val in kwargs.items():
                try:
                    if idfobj[key] == val:
                        return idfobj
                except BadEPFieldError as e:
                    continue


def getidfobjectlist(idf):
    """return a list of all idfobjects in idf"""
    idfobjects = idf.idfobjects
    # idfobjlst = [idfobjects[key] for key in idfobjects if idfobjects[key]]
    idfobjlst = [idfobjects[key] for key in idf.model.dtls if idfobjects[key]]
    # `for key in idf.model.dtls` maintains the order
    # `for key in idfobjects` does not have order
    idfobjlst = itertools.chain.from_iterable(idfobjlst)
    idfobjlst = list(idfobjlst)
    return idfobjlst


def copyidfintoidf(toidf, fromidf):
    """copy fromidf completely into toidf"""
    idfobjlst = getidfobjectlist(fromidf)
    for idfobj in idfobjlst:
        toidf.copyidfobject(idfobj)



# Copyright (c) 2016 Santosh Philip
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================
"""py.test for idf"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from io import StringIO

from eppy.iddcurrent import iddcurrent
from eppy.modeleditor import IDF
from eppy.pytest_helpers import almostequal
import eppy.idf_helpers as idf_helpers

iddfhandle = StringIO(iddcurrent.iddtxt)

if IDF.getiddname() == None:
    IDF.setiddname(iddfhandle)


def test_idfobjectkeys():
    """py.test for idfobjectkeys"""
    expected = [
        "LEAD INPUT",
        "SIMULATION DATA",
        "VERSION",
        "SIMULATIONCONTROL",
        "BUILDING",
    ]
    idf = IDF(StringIO(""))
    result = idf_helpers.idfobjectkeys(idf)
    assert result[:5] == expected


def test_getanymentions():
    """py.test for getanymentions"""
    idf = IDF(StringIO(""))
    mat = idf.newidfobject("MATERIAL", Name="mat")
    aconst = idf.newidfobject("CONSTRUCTION", Name="const")
    foundobjs = idf_helpers.getanymentions(idf, mat)
    assert len(foundobjs) == 1
    assert foundobjs[0] == mat


def test_getobject_use_prevfield():
    """py.test for getobject_use_prevfield"""
    idf = IDF(StringIO(""))
    branch = idf.newidfobject(
        "BRANCH",
        Name="CW Pump Branch",
        Component_1_Object_Type="Pump:VariableSpeed",
        Component_1_Name="CW Circ Pump",
    )
    pump = idf.newidfobject("PUMP:VARIABLESPEED", Name="CW Circ Pump")
    foundobject = idf_helpers.getobject_use_prevfield(idf, branch, "Component_1_Name")
    assert foundobject == pump
    # test for all times it should return None
    foundobject = idf_helpers.getobject_use_prevfield(idf, branch, "Name")
    foundobject = None  # prev field not end with Object_Type
    foundobject = idf_helpers.getobject_use_prevfield(
        idf, branch, "Component_11_Object_Type"
    )
    foundobject = None  # field does not end with "Name"
    foundobject = idf_helpers.getobject_use_prevfield(idf, branch, "Component_3_Name")
    foundobject = None  # bad idfobject key


def test_getidfkeyswithnodes():
    """py.test for getidfkeyswithnodes"""
    nodekeys = idf_helpers.getidfkeyswithnodes()
    # print(len(nodekeys))
    assert "PLANTLOOP" in nodekeys
    assert "ZONE" not in nodekeys


# def test_a():
#     assert 1== 2


def test_getobjectswithnode():
    """py.test for getobjectswithnode"""
    idf = IDF(StringIO(""))
    nodekeys = idf_helpers.getidfkeyswithnodes()
    plantloop = idf.newidfobject(
        "PlantLoop",
        Name="Chilled Water Loop",
        Plant_Side_Inlet_Node_Name="CW Supply Inlet Node",
    )
    branch = idf.newidfobject(
        "Branch",
        Name="CW Pump Branch",
        Component_1_Inlet_Node_Name="CW Supply Inlet Node",
    )
    pump = idf.newidfobject(
        "Pump:VariableSpeed",
        Name="CW Circ Pump",
        Inlet_Node_Name="CW Supply Inlet Node",
    )
    zone = idf.newidfobject("zone")
    foundobjs = idf_helpers.getobjectswithnode(idf, nodekeys, "CW Supply Inlet Node")
    expected = [plantloop, branch, pump]
    expectedset = set([item.key for item in expected])
    resultset = set([item.key for item in foundobjs])
    assert resultset == expectedset
    expectedset = set([item.Name for item in expected])
    resultset = set([item.Name for item in foundobjs])
    assert resultset == expectedset


def test_name2idfobject():
    """py.test for name2idfobject"""
    idf = IDF(StringIO(""))
    plantloopname = "plantloopname"
    branchname = "branchname"
    pumpname = "pumpname"
    zonename = "zonename"
    plantloop = idf.newidfobject(
        "PlantLoop",
        Name=plantloopname,
        Plant_Side_Inlet_Node_Name="CW Supply Inlet Node",
    )
    branch = idf.newidfobject(
        "Branch", Name=branchname, Component_1_Inlet_Node_Name="CW Supply Inlet Node"
    )
    pump = idf.newidfobject(
        "Pump:VariableSpeed", Name=pumpname, Inlet_Node_Name="CW Supply Inlet Node"
    )
    zone = idf.newidfobject("zone", Name=zonename)
    simulation = idf.newidfobject("SimulationControl")
    # - test
    names = [plantloopname, branchname, pumpname, zonename]
    idfobjs = [plantloop, branch, pump, zone]
    for name, idfobj in zip(names, idfobjs):
        result = idf_helpers.name2idfobject(idf, Name=name)
        assert result == idfobj
    # test when objkeys!=None
    objkey = "ZoneHVAC:EquipmentConnections"
    equipconnections = idf.newidfobject(objkey, Zone_Name=zonename)
    result = idf_helpers.name2idfobject(idf, Zone_Name=zonename, objkeys=[objkey])
    assert result == equipconnections


def test_getidfobjectlist():
    """py.test for getidfobjectlist"""
    names = ["a", "b", "c", "d", "e"]
    idf = IDF(StringIO(""))
    idf.newidfobject("building", Name="a")
    idf.newidfobject("building", Name="b")
    idf.newidfobject("Site:Location", Name="c")
    idf.newidfobject("ScheduleTypeLimits", Name="d")
    idf.newidfobject("ScheduleTypeLimits", Name="e")
    result = idf_helpers.getidfobjectlist(idf)
    assert [res.Name for res in result] == names


def test_copyidfintoidf():
    """py.test for copyidfintoidf"""
    tonames = ["a", "b", "c"]
    fromnames = ["d", "e"]
    allnames = ["a", "b", "c", "d", "e"]
    toidf = IDF(StringIO(""))
    toidf.newidfobject("building", Name="a")
    toidf.newidfobject("building", Name="b")
    toidf.newidfobject("Site:Location", Name="c")
    result = idf_helpers.getidfobjectlist(toidf)
    assert [res.Name for res in result] == tonames
    fromidf = IDF(StringIO(""))
    fromidf.newidfobject("ScheduleTypeLimits", Name="d")
    fromidf.newidfobject("ScheduleTypeLimits", Name="e")
    result = idf_helpers.getidfobjectlist(fromidf)
    assert [res.Name for res in result] == fromnames
    idf_helpers.copyidfintoidf(toidf, fromidf)
    result = idf_helpers.getidfobjectlist(toidf)
    assert [res.Name for res in result] == allnames



# Copyright (c) 2012 Santosh Philip
# Copyright (c) 2016 Jamie Bull
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================
"""py.test for loopdiagram.py"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os

import pytest

from eppy.useful_scripts.loopdiagram import clean_edges
from eppy.useful_scripts.loopdiagram import dropnodes
from eppy.useful_scripts.loopdiagram import edges2nodes
from eppy.useful_scripts.loopdiagram import process_idf
from eppy.useful_scripts.loopdiagram import replace_colon
from eppy.pytest_helpers import do_integration_tests


THIS_DIR = os.path.dirname(os.path.abspath(__file__))

RESOURCES_DIR = os.path.join(THIS_DIR, os.pardir, "eppy", "resources")

IDD_FILES = os.path.join(RESOURCES_DIR, "iddfiles")
IDF_FILES = os.path.join(RESOURCES_DIR, "idffiles")


def test_dropnodes():
    """py.test for dropnodes"""
    # test 1
    node = "node"
    (a, b, c, d, e, f, g, h, i) = (
        ("a", node),
        "b",
        ("c", node),
        "d",
        ("e", node),
        "f",
        ("g", node),
        "h",
        ("i", node),
    )
    edges = [(a, b), (b, c), (c, d), (d, e), (e, f), (f, g), (g, h), (h, i)]
    theresult = [("a", "b"), ("b", "d"), ("d", "f"), ("f", "h"), ("h", "i")]
    result = dropnodes(edges)
    assert result == theresult
    # test 2
    (a, b, c, d, e, f, g, h, i, j) = (
        ("a", node),
        "b",
        ("c", node),
        ("d", node),
        "e",
        "f",
        ("g", node),
        ("h", node),
        "i",
        ("j", node),
    )
    edges = [
        (a, b),
        (b, c),
        (c, e),
        (e, g),
        (g, i),
        (i, j),
        (b, d),
        (d, f),
        (f, h),
        (h, i),
    ]
    theresult = [("a", "b"), ("b", "e"), ("e", "i"), ("i", "j"), ("b", "f"), ("f", "i")]
    result = dropnodes(edges)
    assert result == theresult


def test_edges2nodes():
    """py.test for edges2nodes"""
    thedata = (
        ([("a", "b"), ("b", "c"), ("c", "d")], ["a", "b", "c", "d"]),  # edges, nodes
    )
    for edges, nodes in thedata:
        result = edges2nodes(edges)
        assert result == nodes


def test_replace_colon():
    """py.test for replace_colon"""
    data = (("zone:aap", "@", "zone@aap"),)  # s, r, replaced
    for s, r, replaced in data:
        result = replace_colon(s, r)
        assert result == replaced


def test_cleanedges():
    """py.test for cleanedges"""
    data = (
        (
            [("a:a", "a"), (("a", "a"), "a:a"), ("a:a", ("a", "a"))],
            (("a__a", "a"), (("a", "a"), "a__a"), ("a__a", ("a", "a"))),
        ),
        # edg, clean_edg
    )
    for edg, clean_edg in data:
        result = clean_edges(edg)
        assert result == clean_edg


@pytest.mark.skipif(
    not do_integration_tests(), reason="$EPPY_INTEGRATION env var not set"
)
def test_loopdiagram_integration():
    """End-to-end smoke test on an example file"""
    idd = os.path.join(IDD_FILES, "Energy+V8_1_0.idd")
    fname = os.path.join(IDF_FILES, "V8_1_0/Boiler.idf")
    process_idf(fname, idd)



# Copyright (c) 2012 Santosh Philip
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================

"""sample script - autosize all autosizable fields"""
# email to jason
# Jason,
#   Try out the attached script.
# (you might have to change file pathnames)
#
# I don't think you could have written it on your own :-(
# It became apparent to me that eppy needs functions that allows it to look
# into the idd structure for any object and field
# I'll put some thought into this and see how best to write this.
# If that functionality is done well, hopefully anyone should be able to write
# the equivalent code to what I wrote in the attached file.
#
#
# Santosh
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import sys

pathnameto_eplusscripting = "../../"
sys.path.append(pathnameto_eplusscripting)

from eppy import modeleditor
from eppy.modeleditor import IDF
from eppy import bunch_subclass


def autosize_fieldname(idfobject):
    """return autsizeable field names in idfobject"""
    # undocumented stuff in this code
    return [
        fname
        for (fname, dct) in zip(idfobject.objls, idfobject["objidd"])
        if "autosizable" in dct
    ]


iddfile = "../resources/iddfiles/Energy+V8_0_0.idd"
fname1 = "../resources/idffiles/V8_0_0/5ZoneWaterLoopHeatPump.idf"

IDF.setiddname(iddfile)
idf = IDF(fname1)
idf.saveas("./a.idf")


allidfobjects = idf.idfobjects
for objname in list(allidfobjects.keys()):
    idfobjects = allidfobjects[objname]
    for idfobject in idfobjects:
        autofields = autosize_fieldname(idfobject)
        for autofield in autofields:
            idfobject[autofield] = "autosize"

idf.saveas("./b.idf")



"""py.test for simpleread.py"""
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from io import StringIO
import eppy.simpleread as simpleread


def test_idf2txt():
    """py.test for idf2txt"""
    data = (
        (
            """
        VERSION,
            7.3;                      !- Version Identifier

        SIMULATIONCONTROL,
            Yes,                      !- Do Zone Sizing Calculation
            Yes,                      !- Do System Sizing Calculation
            Yes,                      !- Do Plant Sizing Calculation
            No,                       !- Run Simulation for Sizing Periods
            Yes;                      !- Run Simulation for Weather File Run Periods

        BUILDING,
            Empire State Building,    !- Name
            30.0,                     !- North Axis
            City,                     !- Terrain
            0.04,                     !- Loads Convergence Tolerance Value
            0.4,                      !- Temperature Convergence Tolerance Value
            FullExterior,             !- Solar Distribution
            25,                       !- Maximum Number of Warmup Days
            6;                        !- Minimum Number of Warmup Days

        SITE:LOCATION,
            CHICAGO_IL_USA TMY2-94846,    !- Name
            41.78,                    !- Latitude
            -87.75,                   !- Longitude
            -6.0,                     !- Time Zone
            190.0;                    !- Elevation
        """,
            """;

BUILDING,
Empire State Building,
30.0,
City,
0.04,
0.4,
FullExterior,
25.0,
6.0;

SIMULATIONCONTROL,
Yes,
Yes,
Yes,
No,
Yes;

SITE:LOCATION,
CHICAGO_IL_USA TMY2-94846,
41.78,
-87.75,
-6.0,
190.0;

VERSION,
7.3;
""",
        ),  # intxt, outtxt
    )
    for intxt, outtxt in data:
        result = simpleread.idf2txt(intxt)
        assert result == outtxt


def test_idfreadtest():
    """py.test for idfreadtest"""
    data = (
        (
            """!IDD_Version 7.2.0.006
Version,
      \\unique-object
      \\format singleLine
  A1 ; \\field Version Identifier

SimulationControl,
      \\unique-object
  A1, \\field Do Zone Sizing Calculation
  A2, \\field Do System Sizing Calculation
  A3, \\field Do Plant Sizing Calculation
  A4, \\field Run Simulation for Sizing Periods
  A5; \\field Run Simulation for Weather File Run Periods

Building,
       \\unique-object
  A1 , \\field Name
  N1 , \\field North Axis
  A2 , \\field Terrain
  N2 , \\field Loads Convergence Tolerance Value
  N3 , \\field Temperature Convergence Tolerance Value
  A3 , \\field Solar Distribution
  N4 , \\field Maximum Number of Warmup Days
  N5 ; \\field Minimum Number of Warmup Days

Site:Location,
       \\unique-object
  A1 , \\field Name
  N1 , \\field Latitude
  N2 , \\field Longitude
  N3 , \\field Time Zone
  N4 ; \\field Elevation

""",
            """
VERSION,
    7.3;                      !- Version Identifier

SIMULATIONCONTROL,
    Yes,                      !- Do Zone Sizing Calculation
    Yes,                      !- Do System Sizing Calculation
    Yes,                      !- Do Plant Sizing Calculation
    No,                       !- Run Simulation for Sizing Periods
    Yes;                      !- Run Simulation for Weather File Run Periods

BUILDING,
    Empire State Building,    !- Name
    30.0,                     !- North Axis
    City,                     !- Terrain
    0.04,                     !- Loads Convergence Tolerance Value
    0.4,                      !- Temperature Convergence Tolerance Value
    FullExterior,             !- Solar Distribution
    25,                       !- Maximum Number of Warmup Days
    6;                        !- Minimum Number of Warmup Days

SITE:LOCATION,
    CHICAGO_IL_USA TMY2-94846,    !- Name
    41.78,                    !- Latitude
    -87.75,                   !- Longitude
    -6.0,                     !- Time Zone
    190.0;                    !- Elevation
""",
        ),  # iddtxt, idftxt
    )
    for iddtxt, idftxt in data:
        iddhandle = StringIO(iddtxt)
        idfhandle1 = StringIO(idftxt)
        idfhandle2 = StringIO(idftxt)
        result = simpleread.idfreadtest(iddhandle, idfhandle1, idfhandle2)
        assert result == True



# Copyright (c) 2012 Santosh Philip
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================

"""py.test for readhtml.py"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import pytest
import collections
from bs4 import BeautifulSoup
import eppy.results.readhtml as readhtml
from .sample_html import sample_html as SAMPLE_HTML


def test_table2matrix():
    """py.test for table2matrix"""
    thedata = (
        (
            """<table border="1" cellspacing="0" cellpadding="4">
                <tr>
                    <td>1</td>
                    <td>2</td>
                </tr>
                <tr>
                    <td>3</td>
                    <td>4</td>
                </tr>
            </table>""",
            [["1", "2"], ["3", "4"]],
        ),  # tabletxt, rows
        # test when there is crud in the cell of the table
        (
            """<table border="1" cellspacing="0" cellpadding="4">
                <tr>
                    <td>1<aaa>something</aaa></td>
                    <td>2</td>
                </tr>
                <tr>
                    <td>3</td>
                    <td>4</td>
                </tr>
            </table>""",
            [["1", "2"], ["3", "4"]],
        ),  # tabletxt, rows
    )
    for tabletxt, rows in thedata:
        soup = BeautifulSoup(tabletxt, "lxml")
        table = soup.find("table")
        result = readhtml.table2matrix(table)
        assert result == rows


def test_table2val_matrix():
    """py.test for table2val_matrix"""
    thedata = (
        (
            """<table border="1" cellspacing="0" cellpadding="4">
            <tr>
                <td>b</td>
                <td>2</td>
            </tr>
            <tr>
                <td>3</td>
                <td>4</td>
            </tr>
            </table>""",
            [["b", 2], [3, 4]],
        ),  # tabletxt, rows
        # the following test data has a <br> in the <td></td>
        # it will test if tdbr2EOL works correctly
        (
            """<table border="1" cellspacing="0" cellpadding="4">
            <tr>
                <td>b <br> b</td>
                <td>2</td>
            </tr>
            <tr>
                <td>3</td>
                <td>4</td>
            </tr>
            </table>""",
            [["b \n b", 2], [3, 4]],
        ),  # tabletxt, rows
        # has a tag in a cell
        (
            """<table border="1" cellspacing="0" cellpadding="4">
            <tr>
                <td>b<sometag> stuff </sometag></td>
                <td>2</td>
            </tr>
            <tr>
                <td>3</td>
                <td>4</td>
            </tr>
            </table>""",
            [["b", 2], [3, 4]],
        ),  # tabletxt, rows
    )
    for tabletxt, rows in thedata:
        soup = BeautifulSoup(tabletxt, "lxml")
        table = soup.find("table")
        result = readhtml.table2val_matrix(table)
        assert result == rows


def test_gettables():
    """py.test for gettables"""
    thedata = (
        (
            [
                ("Site and Source Energy", [["a", "2"], ["3", "4"]]),
                ("Site to Source Energy Conversion Factors", [["b", "6"], ["7", "8"]]),
                ("Custom Monthly Report", [["c", "16"], ["17", "18"]]),
                ("Custom Monthly Report", [["d", "26"], ["27", "28"]]),
            ],
            False,
        ),  # titlerows, tofloat
        (
            [
                ("Site and Source Energy", [["a", 2], [3, 4]]),
                ("Site to Source Energy Conversion Factors", [["b", 6], [7, 8]]),
                ("Custom Monthly Report", [["c", 16], [17, 18]]),
                ("Custom Monthly Report", [["d", 26], [27, 28]]),
            ],
            True,
        ),  # titlerows, tofloat
    )
    for titlerows, tofloat in thedata:
        result = readhtml.titletable(SAMPLE_HTML, tofloat=tofloat)
        for (title1, rows1), (title2, rows2) in zip(result, titlerows):
            assert title1 == title2
            assert rows1 == rows2
        assert result == titlerows


def test_has_name():
    """py.test for has_name"""
    soup = BeautifulSoup(SAMPLE_HTML, "lxml")
    # soup.p = <p><a href="#toc" style="float: right">Table of Contents</a></p>
    assert readhtml._has_name(soup.p) is True
    # soup.b = <b>EnergyPlus-Windows-OMP-32 7.2.0.006, YMD=2013.01.28 16:38</b>
    assert readhtml._has_name(soup.b) is True
    # soup.p.contents[0] = <a href="#toc" style="float: right">Table of Contents</a>
    assert readhtml._has_name(soup.p.contents[0]) is True
    # soup.b.contents[0] = u'EnergyPlus-Windows-OMP-32 7.2.0.006, YMD=2013.01.28 16:38'
    assert readhtml._has_name(soup.b.contents[0]) is False


def test_lines_table():
    """py.test for lines_table"""
    # soup = BeautifulSoup(SAMPLE_HTML)
    result = readhtml.lines_table(SAMPLE_HTML, False)
    assert result == [
        [
            [
                "Table of Contents",
                "Report: Annual Building Utility Performance Summary",
                "For: Entire Facility",
                "Timestamp: 2014-01-13\n    16:47:19",
                "Values gathered over      8760.00 hours",
                "Site and Source Energy",
            ],
            [["a", "2"], ["3", "4"]],
        ],
        [["Site to Source Energy Conversion Factors"], [["b", "6"], ["7", "8"]]],
        [
            [
                "Report: COMPONENTS OF PEAK ELECTRICAL DEMAND",
                "For: Meter",
                "Timestamp: 2014-01-13\n    16:47:19",
                "Custom Monthly Report",
            ],
            [["c", "16"], ["17", "18"]],
        ],
        [
            [
                "Report: COMPONENTS OF PEAK NET ELECTRICAL DEMAND",
                "For: Meter",
                "Timestamp: 2014-01-13\n    16:47:19",
                "Custom Monthly Report",
            ],
            [["d", "26"], ["27", "28"]],
        ],
    ]


def test_make_ntgrid():
    """py.test make_ntgrid"""
    grid = [
        ["", "a b", "b c", "c d"],
        ["x y", 1, 2, 3],
        ["y z", 4, 5, 6],
        ["z z", 7, 8, 9],
    ]
    result = readhtml._make_ntgrid(grid)
    ntcol = collections.namedtuple("ntcol", "x_y y_z z_z")
    ntrow = collections.namedtuple("ntrow", "a_b b_c c_d")
    assert result == ntcol(
        x_y=ntrow(a_b=1, b_c=2, c_d=3),
        y_z=ntrow(a_b=4, b_c=5, c_d=6),
        z_z=ntrow(a_b=7, b_c=8, c_d=9),
    )


# cell2txt
@pytest.mark.parametrize(
    "tdtxt, expected",
    [
        ('<td align="right">cell text</td>', "cell text"),  # tdtxt, expected
        (
            '<td align="right">cell text<aaa> </aaa></td>',
            "cell text",
        ),  # tdtxt, expected
        (
            '<td align="right">cell text<aaa> </aaa> more text</td>',
            "cell text more text",
        ),  # tdtxt, expected
        (
            '<td align="right">cell text<aaa> something here</aaa> more text</td>',
            "cell text more text",
        ),  # tdtxt, expected
        (
            '<td align="right"><aaa> </aaa>cell text</td>',
            "cell text",
        ),  # tdtxt, expected
        ('<td align="right"></td>', ""),  # tdtxt, expected
    ],
)
def test_cell2txt(tdtxt, expected):
    """py.test for cell2txt"""
    soup = BeautifulSoup(tdtxt, "html.parser")
    td = soup.td
    result = readhtml.cell2txt(td)
    assert result == expected



# Copyright (c) 2016 Santosh Philip
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================
"""py.test for json_functions"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from eppy import modeleditor
from eppy.modeleditor import IDF
from io import StringIO

from eppy.iddcurrent import iddcurrent

from eppy import json_functions


# idd is read only once in this test
# if it has already been read from some other test, it will continue with
# the old reading
iddfhandle = StringIO(iddcurrent.iddtxt)
if IDF.getiddname() == None:
    IDF.setiddname(iddfhandle)


def test_key2elements():
    """py.test for key2elements"""
    data = (
        ("a.b.c.d", ["a", "b", "c", "d"]),  # key, elements
        ("idf.a.Name.name.c", ["idf", "a", "Name.name", "c"]),  # key, elements
        ("idf.a.'Name.name'.c", ["idf", "a", "Name.name", "c"]),  # key, elements
        ("idf.a.'Name.name.n'.c", ["idf", "a", "Name.name.n", "c"]),  # key, elements
    )
    for key, elements in data:
        result = json_functions.key2elements(key)
        # print(result)
        # print(elements)
        assert result == elements


def test_updateidf():
    """py.test for updateidf"""
    iddtxt = """!IDD_Version 8.4.0"""
    data = (
        (
            """Version,
        8.3;                     !- Version Identifier

    """,
            {"idf.version..Version_Identifier": "0.1"},
            "version",
            "Version_Identifier",
            "0.1",
        ),  # idftxt, dct, key, field, fieldval
        (
            """SimulationControl,
    No,                      !- Do Zone Sizing Calculation
    No,                      !- Do System Sizing Calculation
    No,                      !- Do Plant Sizing Calculation
    No,                      !- Run Simulation for Sizing Periods
    Yes;                     !- Run Simulation for Weather File Run Periods
    """,
            {"idf.SimulationControl..Do_Zone_Sizing_Calculation": "Yes"},
            "SimulationControl",
            "Do_Zone_Sizing_Calculation",
            "Yes",
        ),  # idftxt, dct, key, field, fieldval
        (
            """Building,
    Untitled,                !- Name
    0.0,                     !- North Axis {deg}
    City,                    !- Terrain
    0.04,                    !- Loads Convergence Tolerance Value
    0.4,                     !- Temperature Convergence Tolerance Value {deltaC}
    FullInteriorAndExterior, !- Solar Distribution
    25,                      !- Maximum Number of Warmup Days
    ;                        !- Minimum Number of Warmup Days
    """,
            {"idf.BUilding.Untitled.Terrain": "Rural"},
            "Building",
            "Terrain",
            "Rural",
        ),  # idftxt, dct, key, field, fieldval
        # make a new object
        (
            """
    """,
            {"idf.BUilding.Taj.Terrain": "Rural"},
            "Building",
            "Terrain",
            "Rural",
        ),  # idftxt, dct, key, field, fieldval
        # make a new object with no Name field
        (
            """
    """,
            {"idf.GlobalGeometryRules..Starting_Vertex_Position": "UpperLeftCorner"},
            "GlobalGeometryRules",
            "Starting_Vertex_Position",
            "UpperLeftCorner",
        ),  # idftxt, dct, key, field, fieldval
        (
            """Building,
    Name.name,                !- Name
    0.0,                     !- North Axis {deg}
    City,                    !- Terrain
    0.04,                    !- Loads Convergence Tolerance Value
    0.4,                     !- Temperature Convergence Tolerance Value {deltaC}
    FullInteriorAndExterior, !- Solar Distribution
    25,                      !- Maximum Number of Warmup Days
    ;                        !- Minimum Number of Warmup Days
    """,
            {"idf.BUilding.Name.name.Terrain": "Rural"},
            "Building",
            "Terrain",
            "Rural",
        ),  # idftxt, dct, key, field, fieldval
        (
            """Building,
    Name.name,                !- Name
    0.0,                     !- North Axis {deg}
    City,                    !- Terrain
    0.04,                    !- Loads Convergence Tolerance Value
    0.4,                     !- Temperature Convergence Tolerance Value {deltaC}
    FullInteriorAndExterior, !- Solar Distribution
    25,                      !- Maximum Number of Warmup Days
    ;                        !- Minimum Number of Warmup Days
    """,
            {"idf.BUilding.'Name.name'.Terrain": "Rural"},
            "Building",
            "Terrain",
            "Rural",
        ),  # idftxt, dct, key, field, fieldval
    )
    for idftxt, dct, key, field, fieldval in data:
        idfhandle = StringIO(idftxt)
        idf = IDF(idfhandle)
        json_functions.updateidf(idf, dct)
        assert idf.idfobjects[key][0][field] == fieldval



# Copyright (c) 2012 Santosh Philip
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================

"""py.test for hvacbuilder"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from io import StringIO

import eppy.hvacbuilder as hvacbuilder
from eppy.iddcurrent import iddcurrent
from eppy.modeleditor import IDF


# idd is read only once in this test
# if it has already been read from some other test, it will continue with the old reading
iddfhandle = StringIO(iddcurrent.iddtxt)
if IDF.getiddname() == None:
    IDF.setiddname(iddfhandle)


def test_flattencopy():
    """py.test for flattencopy"""
    tdata = (
        ([1, 2], [1, 2]),  # lst , nlst
        ([1, 2, [3, 4]], [1, 2, 3, 4]),  # lst , nlst
        ([1, 2, [3, [4, 5, 6], 7, 8]], [1, 2, 3, 4, 5, 6, 7, 8]),  # lst , nlst
        ([1, 2, [3, [4, 5, [6, 7], 8], 9]], [1, 2, 3, 4, 5, 6, 7, 8, 9]),
        # lst , nlst
    )
    for lst, nlst in tdata:
        result = hvacbuilder.flattencopy(lst)
        assert result == nlst


def test_makeplantloop():
    """pytest for makeplantloop"""
    tdata = (
        (
            "",
            "p_loop",
            ["sb0", ["sb1", "sb2", "sb3"], "sb4"],
            ["db0", ["db1", "db2", "db3"], "db4"],
            """BRANCH, sb0, 0.0, , Pipe:Adiabatic, sb0_pipe, p_loop Supply Inlet,
        sb0_pipe_outlet, Bypass;BRANCH, sb1, 0.0, , Pipe:Adiabatic, sb1_pipe,
        sb1_pipe_inlet, sb1_pipe_outlet, Bypass;BRANCH, sb2, 0.0, ,
        Pipe:Adiabatic, sb2_pipe, sb2_pipe_inlet, sb2_pipe_outlet,
        Bypass;BRANCH, sb3, 0.0, , Pipe:Adiabatic, sb3_pipe, sb3_pipe_inlet,
        sb3_pipe_outlet, Bypass;BRANCH, sb4, 0.0, , Pipe:Adiabatic, sb4_pipe,
        sb4_pipe_inlet, p_loop Supply Outlet, Bypass;BRANCH, db0, 0.0, ,
        Pipe:Adiabatic, db0_pipe, p_loop Demand Inlet, db0_pipe_outlet,
        Bypass;BRANCH, db1, 0.0, , Pipe:Adiabatic, db1_pipe, db1_pipe_inlet,
        db1_pipe_outlet, Bypass;BRANCH, db2, 0.0, , Pipe:Adiabatic, db2_pipe,
        db2_pipe_inlet, db2_pipe_outlet, Bypass;BRANCH, db3, 0.0, ,
        Pipe:Adiabatic, db3_pipe, db3_pipe_inlet, db3_pipe_outlet,
        Bypass;BRANCH, db4, 0.0, , Pipe:Adiabatic, db4_pipe, db4_pipe_inlet,
        p_loop Demand Outlet, Bypass;BRANCHLIST, p_loop Supply Branchs,
        sb0, sb1, sb2, sb3, sb4;BRANCHLIST, p_loop Demand Branchs, db0,
        db1, db2, db3, db4;CONNECTOR:SPLITTER, p_loop_supply_splitter,
        sb0, sb1, sb2, sb3;CONNECTOR:SPLITTER, p_loop_demand_splitter,
        db0, db1, db2, db3;CONNECTOR:MIXER, p_loop_supply_mixer, sb4,
        sb1, sb2, sb3;CONNECTOR:MIXER, p_loop_demand_mixer, db4, db1,
        db2, db3;CONNECTORLIST, p_loop Supply Connectors, Connector:Splitter,
        p_loop_supply_splitter, Connector:Mixer,
        p_loop_supply_mixer;CONNECTORLIST, p_loop Demand Connectors,
        Connector:Splitter, p_loop_demand_splitter, Connector:Mixer,
        p_loop_demand_mixer;PIPE:ADIABATIC, sb0_pipe, p_loop Supply Inlet,
        sb0_pipe_outlet;PIPE:ADIABATIC, sb1_pipe, sb1_pipe_inlet,
        sb1_pipe_outlet;PIPE:ADIABATIC, sb2_pipe, sb2_pipe_inlet,
        sb2_pipe_outlet;PIPE:ADIABATIC, sb3_pipe, sb3_pipe_inlet,
        sb3_pipe_outlet;PIPE:ADIABATIC, sb4_pipe, sb4_pipe_inlet,
        p_loop Supply Outlet;PIPE:ADIABATIC, db0_pipe,
        p_loop Demand Inlet, db0_pipe_outlet;PIPE:ADIABATIC, db1_pipe,
        db1_pipe_inlet, db1_pipe_outlet;PIPE:ADIABATIC, db2_pipe,
        db2_pipe_inlet, db2_pipe_outlet;PIPE:ADIABATIC, db3_pipe,
        db3_pipe_inlet, db3_pipe_outlet;PIPE:ADIABATIC, db4_pipe,
        db4_pipe_inlet, p_loop Demand Outlet;PLANTLOOP, p_loop, Water, , ,
        , , , , 0.0, Autocalculate, p_loop Supply Inlet,
        p_loop Supply Outlet, p_loop Supply Branchs,
        p_loop Supply Connectors, p_loop Demand Inlet, p_loop Demand Outlet,
        p_loop Demand Branchs, p_loop Demand Connectors, Sequential, ,
        SingleSetpoint, None, None;""",
        ),  # blankidf, loopname, sloop, dloop, nidf
    )
    for blankidf, loopname, sloop, dloop, nidf in tdata:
        fhandle = StringIO("")
        idf1 = IDF(fhandle)
        loopname = "p_loop"
        sloop = ["sb0", ["sb1", "sb2", "sb3"], "sb4"]
        dloop = ["db0", ["db1", "db2", "db3"], "db4"]
        hvacbuilder.makeplantloop(idf1, loopname, sloop, dloop)
        idf2 = IDF(StringIO(nidf))
        # print('=' * 15)
        # print(idf1.model)
        # print('-' * 15)
        # print(idf2.model)
        # print('=' * 15)
        assert str(idf1.model) == str(idf2.model)


def test_makecondenserloop():
    """pytest for makecondenserloop"""
    tdata = (
        (
            "",
            "c_loop",
            ["sb0", ["sb1", "sb2", "sb3"], "sb4"],
            ["db0", ["db1", "db2", "db3"], "db4"],
            """BRANCH, sb0, 0.0, , Pipe:Adiabatic, sb0_pipe,
        c_loop Cond_Supply Inlet, sb0_pipe_outlet, Bypass;  BRANCH, sb1, 0.0,
        , Pipe:Adiabatic, sb1_pipe, sb1_pipe_inlet, sb1_pipe_outlet,
        Bypass;  BRANCH, sb2, 0.0, , Pipe:Adiabatic, sb2_pipe,
        sb2_pipe_inlet, sb2_pipe_outlet, Bypass;  BRANCH, sb3, 0.0, ,
        Pipe:Adiabatic, sb3_pipe, sb3_pipe_inlet, sb3_pipe_outlet,
        Bypass;  BRANCH, sb4, 0.0, , Pipe:Adiabatic, sb4_pipe,
        sb4_pipe_inlet, c_loop Cond_Supply Outlet, Bypass;  BRANCH,
        db0, 0.0, , Pipe:Adiabatic, db0_pipe, c_loop Demand Inlet,
        db0_pipe_outlet, Bypass;  BRANCH, db1, 0.0, , Pipe:Adiabatic, db1_pipe,
        db1_pipe_inlet, db1_pipe_outlet, Bypass;  BRANCH, db2, 0.0, ,
        Pipe:Adiabatic, db2_pipe, db2_pipe_inlet, db2_pipe_outlet, Bypass;
        BRANCH, db3, 0.0, , Pipe:Adiabatic, db3_pipe, db3_pipe_inlet,
        db3_pipe_outlet, Bypass;  BRANCH, db4, 0.0, , Pipe:Adiabatic,
        db4_pipe, db4_pipe_inlet, c_loop Demand Outlet, Bypass;
        BRANCHLIST, c_loop Cond_Supply Branchs, sb0, sb1, sb2, sb3, sb4;
        BRANCHLIST, c_loop Condenser Demand Branchs, db0, db1, db2, db3,
        db4;  CONNECTOR:SPLITTER, c_loop_supply_splitter, sb0, sb1,
        sb2, sb3;  CONNECTOR:SPLITTER, c_loop_demand_splitter, db0, db1, db2,
        db3;  CONNECTOR:MIXER, c_loop_supply_mixer, sb4, sb1, sb2, sb3;
        CONNECTOR:MIXER, c_loop_demand_mixer, db4, db1, db2, db3;
        CONNECTORLIST, c_loop Cond_Supply Connectors, Connector:Splitter,
        c_loop_supply_splitter, Connector:Mixer, c_loop_supply_mixer;
        CONNECTORLIST, c_loop Condenser Demand Connectors,
        Connector:Splitter, c_loop_demand_splitter, Connector:Mixer,
        c_loop_demand_mixer;  PIPE:ADIABATIC, sb0_pipe,
        c_loop Cond_Supply Inlet, sb0_pipe_outlet;  PIPE:ADIABATIC,
        sb1_pipe, sb1_pipe_inlet, sb1_pipe_outlet;  PIPE:ADIABATIC, sb2_pipe,
        sb2_pipe_inlet, sb2_pipe_outlet;  PIPE:ADIABATIC, sb3_pipe,
        sb3_pipe_inlet, sb3_pipe_outlet;  PIPE:ADIABATIC, sb4_pipe,
        sb4_pipe_inlet, c_loop Cond_Supply Outlet;  PIPE:ADIABATIC,
        db0_pipe, c_loop Demand Inlet, db0_pipe_outlet;  PIPE:ADIABATIC,
        db1_pipe, db1_pipe_inlet, db1_pipe_outlet;  PIPE:ADIABATIC,
        db2_pipe, db2_pipe_inlet, db2_pipe_outlet;  PIPE:ADIABATIC,
        db3_pipe, db3_pipe_inlet, db3_pipe_outlet;  PIPE:ADIABATIC, db4_pipe,
        db4_pipe_inlet, c_loop Demand Outlet;  CONDENSERLOOP, c_loop, Water, ,
        , , , , , 0.0, Autocalculate, c_loop Cond_Supply Inlet,
        c_loop Cond_Supply Outlet, c_loop Cond_Supply Branchs,
        c_loop Cond_Supply Connectors, c_loop Demand Inlet,
        c_loop Demand Outlet, c_loop Condenser Demand Branchs,
        c_loop Condenser Demand Connectors, Sequential, None;  """,
        ),  # blankidf, loopname, sloop, dloop, nidf
    )
    for blankidf, loopname, sloop, dloop, nidf in tdata:

        fhandle = StringIO("")
        idf1 = IDF(fhandle)
        loopname = "c_loop"
        sloop = ["sb0", ["sb1", "sb2", "sb3"], "sb4"]
        dloop = ["db0", ["db1", "db2", "db3"], "db4"]
        hvacbuilder.makecondenserloop(idf1, loopname, sloop, dloop)
        idf2 = IDF(StringIO(nidf))
        assert str(idf1.model) == str(idf2.model)


def test_getbranchcomponents():
    """py.test for getbranchcomponents"""
    tdata = (
        (
            """BRANCH,
            sb1,
            0.0,
            ,
            PIPE:ADIABATIC,
            np1,
            np1_inlet,
            np1_np2_node,
            ,
            PIPE:ADIABATIC,
            np2,
            np1_np2_node,
            np2_outlet,
            ;
            """,
            True,
            [("PIPE:ADIABATIC", "np1"), ("PIPE:ADIABATIC", "np2")],
        ),  # idftxt, utest, componentlist
        (
            """BRANCH,
            sb1,
            0.0,
            ,
            PIPE:ADIABATIC,
            np1,
            np1_inlet,
            np1_np2_node,
            ,
            PIPE:ADIABATIC,
            np2,
            np1_np2_node,
            np2_outlet,
            ;
            PIPE:ADIABATIC,
            np1,
            np1_inlet,
            np1_np2_node;

            PIPE:ADIABATIC,
            np2,
            np1_np2_node,
            np2_outlet;

            """,
            False,
            [
                ["PIPE:ADIABATIC", "np1", "np1_inlet", "np1_np2_node"],
                ["PIPE:ADIABATIC", "np2", "np1_np2_node", "np2_outlet"],
            ],
        ),
        # idftxt, utest, componentlist
    )
    for idftxt, utest, componentlist in tdata:
        fhandle = StringIO(idftxt)
        idf = IDF(fhandle)
        branch = idf.idfobjects["BRANCH"][0]
        result = hvacbuilder.getbranchcomponents(idf, branch, utest=utest)
        if utest:
            assert result == componentlist
        else:
            lresult = [item.obj for item in result]
            assert lresult == componentlist


def test_renamenodes():
    """py.test for renamenodes"""
    idftxt = """PIPE:ADIABATIC,
         np1,
         np1_inlet,
         np1_outlet;
         !- ['np1_outlet', 'np1_np2_node'];

    BRANCH,
         sb0,
         0.0,
         ,
         Pipe:Adiabatic,
         np1,
         np1_inlet,
         np1_outlet,
         Bypass;
    """
    outtxt = """PIPE:ADIABATIC,
         np1,
         np1_inlet,
         np1_np2_node;
         !- ['np1_outlet', 'np1_np2_node'];

    BRANCH,
         sb0,
         0.0,
         ,
         Pipe:Adiabatic,
         np1,
         np1_inlet,
         np1_np2_node,
         Bypass;
    """
    # !- ['np1_outlet', 'np1_np2_node'];
    fhandle = StringIO(idftxt)
    idf = IDF(fhandle)
    pipe = idf.idfobjects["PIPE:ADIABATIC"][0]
    pipe.Outlet_Node_Name = [
        "np1_outlet",
        "np1_np2_node",
    ]  # this is the first step of the replace
    hvacbuilder.renamenodes(idf, fieldtype="node")
    outidf = IDF(StringIO(outtxt))
    result = idf.idfobjects["PIPE:ADIABATIC"][0].obj
    assert result == outidf.idfobjects["PIPE:ADIABATIC"][0].obj


def test_getfieldnamesendswith():
    """py.test for getfieldnamesendswith"""
    idftxt = """PIPE:ADIABATIC,
        np2,                      !- Name
        np1_np2_node,             !- Inlet Node Name
        np2_outlet;               !- Outlet Node Name

    """
    tdata = (
        ("Inlet_Node_Name", ["Inlet_Node_Name"]),  # endswith, fieldnames
        ("Node_Name", ["Inlet_Node_Name", "Outlet_Node_Name"]),  # endswith, fieldnames
        (
            "Name",
            ["Name", "Inlet_Node_Name", "Outlet_Node_Name"],
        ),  # endswith, fieldnames
    )
    fhandle = StringIO(idftxt)
    idf = IDF(fhandle)
    idfobject = idf.idfobjects["PIPE:ADIABATIC"][0]
    for endswith, fieldnames in tdata:
        result = hvacbuilder.getfieldnamesendswith(idfobject, endswith)
        assert result == fieldnames


def test_getnodefieldname():
    """py.test for getnodefieldname"""
    tdata = (
        ("PIPE:ADIABATIC", "pipe1", "Inlet_Node_Name", "", "Inlet_Node_Name"),
        # objtype, objname, endswith, fluid, nodefieldname
        (
            "CHILLER:ELECTRIC",
            "pipe1",
            "Inlet_Node_Name",
            "",
            "Chilled_Water_Inlet_Node_Name",
        ),
        # objtype, objname, endswith, fluid, nodefieldname
        (
            "COIL:COOLING:WATER",
            "pipe1",
            "Inlet_Node_Name",
            "Water",
            "Water_Inlet_Node_Name",
        ),
        # objtype, objname, endswith, fluid, nodefieldname
        (
            "COIL:COOLING:WATER",
            "pipe1",
            "Inlet_Node_Name",
            "Air",
            "Air_Inlet_Node_Name",
        ),
        # objtype, objname, endswith, fluid, nodefieldname
        (
            "COIL:COOLING:WATER",
            "pipe1",
            "Outlet_Node_Name",
            "Air",
            "Air_Outlet_Node_Name",
        ),
        # objtype, objname, endswith, fluid, nodefieldname
    )
    for objtype, objname, endswith, fluid, nodefieldname in tdata:
        fhandle = StringIO("")
        idf = IDF(fhandle)
        idfobject = idf.newidfobject(objtype, Name=objname)
        result = hvacbuilder.getnodefieldname(idfobject, endswith, fluid)
        assert result == nodefieldname


def test_connectcomponents():
    """py.test for connectcomponents"""
    fhandle = StringIO("")
    idf = IDF(fhandle)

    tdata = (
        (
            [
                (idf.newidfobject("PIPE:ADIABATIC", Name="pipe1"), None),
                (idf.newidfobject("PIPE:ADIABATIC", Name="pipe2"), None),
            ],
            ["pipe1_Inlet_Node_Name", ["pipe2_Inlet_Node_Name", "pipe1_pipe2_node"]],
            [["pipe1_Outlet_Node_Name", "pipe1_pipe2_node"], "pipe2_Outlet_Node_Name"],
            "",
        ),
        # components_thisnodes, inlets, outlets, fluid
        (
            [
                (idf.newidfobject("Coil:Cooling:Water", Name="pipe1"), "Water_"),
                (idf.newidfobject("Coil:Cooling:Water", Name="pipe2"), "Water_"),
            ],
            [
                "pipe1_Water_Inlet_Node_Name",
                "",
                "pipe2_Water_Inlet_Node_Name",
                ["", "pipe1_pipe2_node"],
            ],
            [
                ["pipe1_Water_Outlet_Node_Name", "pipe1_pipe2_node"],
                "",
                "pipe2_Water_Outlet_Node_Name",
                "",
            ],
            "Air",
        ),
        # components_thisnodes, inlets, outlets, fluid
        (
            [
                (idf.newidfobject("PIPE:ADIABATIC", Name="pipe1"), None),
                (idf.newidfobject("Coil:Cooling:Water", Name="pipe2"), "Water_"),
            ],
            [
                "pipe1_Inlet_Node_Name",
                "pipe2_Water_Inlet_Node_Name",
                ["pipe2_Air_Inlet_Node_Name", "pipe1_pipe2_node"],
            ],
            [
                ["pipe1_Outlet_Node_Name", "pipe1_pipe2_node"],
                "pipe2_Water_Outlet_Node_Name",
                "",
            ],
            "Air",
        ),
        # components_thisnodes, inlets, outlets, fluid
    )
    for components_thisnodes, inlets, outlets, fluid in tdata:
        # init the nodes in the new components
        for component, thisnode in components_thisnodes:
            hvacbuilder.initinletoutlet(idf, component, thisnode)
        hvacbuilder.connectcomponents(idf, components_thisnodes, fluid)
        inresult = []
        for component, thisnode in components_thisnodes:
            fldnames = hvacbuilder.getfieldnamesendswith(component, "Inlet_Node_Name")
            for name in fldnames:
                inresult.append(component[name])
        assert inresult == inresult
        outresult = []
        for component, thisnode in components_thisnodes:
            fldnames = hvacbuilder.getfieldnamesendswith(component, "Outlet_Node_Name")
            for name in fldnames:
                outresult.append(component[name])
        assert outresult == outlets


def test_initinletoutlet():
    """py.test for initinletoutlet"""
    tdata = (
        (
            "PIPE:ADIABATIC",
            "apipe",
            None,
            True,
            ["apipe_Inlet_Node_Name"],
            ["apipe_Outlet_Node_Name"],
        ),
        # idfobjectkey, idfobjname, thisnode, force, inlets, outlets
        ("PIPE:ADIABATIC", "apipe", None, False, ["Gumby"], ["apipe_Outlet_Node_Name"]),
        # idfobjectkey, idfobjname, thisnode, force, inlets, outlets
        (
            "Coil:Cooling:Water",
            "acoil",
            "Water_",
            True,
            ["acoil_Water_Inlet_Node_Name", ""],
            ["acoil_Water_Outlet_Node_Name", ""],
        ),
        # idfobjectkey, idfobjname, thisnode, force, inlets, outlets
    )
    fhandle = StringIO("")
    idf = IDF(fhandle)
    for idfobjectkey, idfobjname, thisnode, force, inlets, outlets in tdata:
        idfobject = idf.newidfobject(idfobjectkey, Name=idfobjname)
        inodefields = hvacbuilder.getfieldnamesendswith(idfobject, "Inlet_Node_Name")
        idfobject[inodefields[0]] = "Gumby"
        hvacbuilder.initinletoutlet(idf, idfobject, thisnode, force=force)
        inodefields = hvacbuilder.getfieldnamesendswith(idfobject, "Inlet_Node_Name")
        for nodefield, inlet in zip(inodefields, inlets):
            result = idfobject[nodefield]
            assert result == inlet
        onodefields = hvacbuilder.getfieldnamesendswith(idfobject, "Outlet_Node_Name")
        for nodefield, outlet in zip(onodefields, outlets):
            result = idfobject[nodefield]
            assert result == outlet


def test_componentsintobranch():
    """py.test for componentsintobranch"""
    tdata = (
        (
            """BRANCH,
             sb0,
             0.0,
             ,
             Pipe:Adiabatic,
             sb0_pipe,
             p_loop Supply Inlet,
             sb0_pipe_outlet,
             Bypass;
             """,
            [("PIPE:ADIABATIC", "pipe1", None), ("PIPE:ADIABATIC", "pipe2", None)],
            "",
            [
                "PIPE:ADIABATIC",
                "pipe1",
                "pipe1_Inlet_Node_Name",
                "pipe1_Outlet_Node_Name",
                "",
                "PIPE:ADIABATIC",
                "pipe2",
                "pipe2_Inlet_Node_Name",
                "pipe2_Outlet_Node_Name",
                "",
            ],
        ),
        # idftxt, complst, fluid, branchcomps
        (
            """BRANCH,
            sb0,
            0.0,
            ,
            Pipe:Adiabatic,
            sb0_pipe,
            p_loop Supply Inlet,
            sb0_pipe_outlet,
            Bypass;
            """,
            [
                ("PIPE:ADIABATIC", "pipe1", None),
                ("CHILLER:ELECTRIC", "chiller", "Chilled_Water_"),
            ],
            "",
            [
                "PIPE:ADIABATIC",
                "pipe1",
                "pipe1_Inlet_Node_Name",
                "pipe1_Outlet_Node_Name",
                "",
                "CHILLER:ELECTRIC",
                "chiller",
                "chiller_Chilled_Water_Inlet_Node_Name",
                "chiller_Chilled_Water_Outlet_Node_Name",
                "",
            ],
        ),
        # idftxt, complst, fluid, branchcomps
    )
    for ii, (idftxt, complst, fluid, branchcomps) in enumerate(tdata):
        fhandle = StringIO(idftxt)
        idf = IDF(fhandle)
        components_thisnodes = [
            (idf.newidfobject(key, Name=nm), thisnode) for key, nm, thisnode in complst
        ]
        fnc = hvacbuilder.initinletoutlet
        components_thisnodes = [
            (fnc(idf, cp, thisnode), thisnode) for cp, thisnode in components_thisnodes
        ]
        branch = idf.idfobjects["BRANCH"][0]
        branch = hvacbuilder.componentsintobranch(
            idf, branch, components_thisnodes, fluid
        )
        assert branch.obj[4:] == branchcomps


def test_replacebranch():
    """py.test for replacebranch"""
    tdata = (
        (
            "p_loop",
            ["sb0", ["sb1", "sb2", "sb3"], "sb4"],
            ["db0", ["db1", "db2", "db3"], "db4"],
            "sb0",
            [
                ("Chiller:Electric", "Central_Chiller", "Chilled_Water_"),
                ("PIPE:ADIABATIC", "np1", None),
                ("PIPE:ADIABATIC", "np2", None),
            ],
            "Water",
            [
                "BRANCH",
                "sb0",
                0.0,
                "",
                "CHILLER:ELECTRIC",
                "Central_Chiller",
                "p_loop Supply Inlet",
                "Central_Chiller_np1_node",
                "",
                "PIPE:ADIABATIC",
                "np1",
                "Central_Chiller_np1_node",
                "np1_np2_node",
                "",
                "PIPE:ADIABATIC",
                "np2",
                "np1_np2_node",
                "np2_Outlet_Node_Name",
                "",
            ],
        ),  # loopname, sloop, dloop, branchname, componenttuple, fluid, outbranch
    )
    for (loopname, sloop, dloop, branchname, componenttuple, fluid, outbranch) in tdata:
        fhandle = StringIO("")
        idf = IDF(fhandle)
        loop = hvacbuilder.makeplantloop(idf, loopname, sloop, dloop)
        components_thisnodes = [
            (idf.newidfobject(key, Name=nm), thisnode)
            for key, nm, thisnode in componenttuple
        ]
        branch = idf.getobject("BRANCH", branchname)
        newbr = hvacbuilder.replacebranch(
            idf, loop, branch, components_thisnodes, fluid=fluid
        )
        assert newbr.obj == outbranch


def test_makepipecomponent():
    """py.test for makepipecomponent"""
    tdata = (
        (
            "apipe",
            ["PIPE:ADIABATIC", "apipe", "apipe_inlet", "apipe_outlet"],
        ),  # pname, pipe_obj
        (
            "bpipe",
            ["PIPE:ADIABATIC", "bpipe", "bpipe_inlet", "bpipe_outlet"],
        ),  # pname, pipe_obj
    )
    for pname, pipe_obj in tdata:
        fhandle = StringIO("")
        idf = IDF(fhandle)
        result = hvacbuilder.makepipecomponent(idf, pname)
        assert result.obj == pipe_obj


def test_makeductcomponent():
    """py.test for makeductcomponent"""
    tdata = (
        ("aduct", ["DUCT", "aduct", "aduct_inlet", "aduct_outlet"]),  # dname, duct_obj
    )
    for dname, duct_obj in tdata:
        fhandle = StringIO("")
        idf = IDF(fhandle)
        result = hvacbuilder.makeductcomponent(idf, dname)
        assert result.obj == duct_obj


def test_makepipebranch():
    """py.test for makepipebranch"""
    tdata = (
        (
            "p_branch",
            [
                "BRANCH",
                "p_branch",
                0.0,
                "",
                "Pipe:Adiabatic",
                "p_branch_pipe",
                "p_branch_pipe_inlet",
                "p_branch_pipe_outlet",
                "Bypass",
            ],
            [
                "PIPE:ADIABATIC",
                "p_branch_pipe",
                "p_branch_pipe_inlet",
                "p_branch_pipe_outlet",
            ],
        ),  # pb_name, branch_obj, pipe_obj
    )
    for pb_name, branch_obj, pipe_obj in tdata:
        fhandle = StringIO("")
        idf = IDF(fhandle)
        result = hvacbuilder.makepipebranch(idf, pb_name)
        assert result.obj == branch_obj
        thepipe = idf.getobject("PIPE:ADIABATIC", result.Component_1_Name)
        assert thepipe.obj == pipe_obj


def test_makeductbranch():
    """py.test for makeductbranch"""
    tdata = (
        (
            "d_branch",
            [
                "BRANCH",
                "d_branch",
                0.0,
                "",
                "duct",
                "d_branch_duct",
                "d_branch_duct_inlet",
                "d_branch_duct_outlet",
                "Bypass",
            ],
            ["DUCT", "d_branch_duct", "d_branch_duct_inlet", "d_branch_duct_outlet"],
        ),  # db_name, branch_obj, duct_obj
    )
    for db_name, branch_obj, duct_obj in tdata:
        fhandle = StringIO("")
        idf = IDF(fhandle)
        result = hvacbuilder.makeductbranch(idf, db_name)
        assert result.obj == branch_obj
        theduct = idf.getobject("DUCT", result.Component_1_Name)
        assert theduct.obj == duct_obj


def test_clean_listofcomponents():
    """py.test for _clean_listofcomponents"""
    data = (
        ([1, 2], [(1, None), (2, None)]),  # lst, clst
        ([(1, None), 2], [(1, None), (2, None)]),  # lst, clst
        ([(1, "stuff"), 2], [(1, "stuff"), (2, None)]),  # lst, clst
    )
    for lst, clst in data:
        result = hvacbuilder._clean_listofcomponents(lst)
        assert result == clst


def test_clean_listofcomponents_tuples():
    """py.test for _clean_listofcomponents_tuples"""
    data = (
        ([(1, 2), (2, 3)], [(1, 2, None), (2, 3, None)]),  # lst, clst
        ([(1, 2, None), (2, 3)], [(1, 2, None), (2, 3, None)]),  # lst, clst
        ([(1, 2, "stuff"), (2, 3)], [(1, 2, "stuff"), (2, 3, None)]),  # lst, clst
    )
    for lst, clst in data:
        result = hvacbuilder._clean_listofcomponents_tuples(lst)
        assert result == clst



# Copyright (c) 2012 Santosh Philip
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================


"""
Do a diff between two idf files.
Prints the diff in csv  or html file format.
You can redirect the output to a file and open the file using as a spreadsheet or by using a browser
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import argparse

from pprint import pprint
import sys
import itertools

try:
    from itertools import zip_longest as zip_longest
except:
    from itertools import izip_longest as zip_longest
from bs4 import BeautifulSoup, Tag

pathnameto_eplusscripting = "../../"
sys.path.append(pathnameto_eplusscripting)

from eppy.bunch_subclass import BadEPFieldError
from eppy.modeleditor import IDF
from eppy.easyopen import easyopen
from eppy.modeleditor import IDDAlreadySetError

help_message = """
The help message goes here.
"""


class IDDMismatchError(Exception):
    pass


class Usage(Exception):
    def __init__(self, msg):
        self.msg = msg


def getobjname(item):
    """return obj name or blank """
    try:
        objname = item.Name
    except BadEPFieldError as e:
        objname = " "
    return objname


def theheader(n1, n2):
    """return the csv header"""
    s = "Object Key, Object Name, Field Name, %s, %s" % ("file1", "file2")
    return s.split(",")


class DtlsSorter(object):
    """helps me to sort it using the order of keys in idd file"""

    def __init__(self, dtls):
        self.dtlsorder = {j: i for i, j in enumerate(dtls)}

    def getkey(self, item):
        return self.dtlsorder[item[0]]  # item[0] is the object key


def makecsvdiffs(thediffs, idf1, idf2):
    """return the csv to be displayed"""
    dtls = idf1.model.dtls  # undocumented variable
    return makecsvdiffs_raw(thediffs, dtls, idf1.idfname, idf2.idfname)


def makecsvdiffs_raw(thediffs, dtls, n1, n2):
    """return the csv to be displayed - the args here are tricky
    This function is called by makecsvdiffs.
    Best not to call directly"""

    def ishere(val):
        if val == None:
            return "not here"
        else:
            return "is here"

    rows = []
    rows.append(["file1 = %s" % (n1,)])
    rows.append(["file2 = %s" % (n2,)])
    rows.append("")
    rows.append(theheader(n1, n2))
    keys = list(thediffs.keys())  # ensures sorting by Name
    keys.sort()
    # sort the keys in the same order as in the idd
    dtlssorter = DtlsSorter(dtls)
    keys = sorted(keys, key=dtlssorter.getkey)
    for key in keys:
        if len(key) == 2:
            rw2 = [""] + [ishere(i) for i in thediffs[key]]
        else:
            rw2 = list(thediffs[key])
        rw1 = list(key)
        rows.append(rw1 + rw2)
    return rows


def idfdiffs(idf1, idf2):
    """return the diffs between the two idfs"""
    # for any object type, it is sorted by name
    thediffs = {}
    keys = idf1.model.dtls  # undocumented variable

    for akey in keys:
        idfobjs1 = idf1.idfobjects[akey]
        idfobjs2 = idf2.idfobjects[akey]
        names = set(
            [getobjname(i) for i in idfobjs1] + [getobjname(i) for i in idfobjs2]
        )
        names = sorted(names)
        idfobjs1 = sorted(idfobjs1, key=lambda idfobj: idfobj["obj"])
        idfobjs2 = sorted(idfobjs2, key=lambda idfobj: idfobj["obj"])
        for name in names:
            n_idfobjs1 = [item for item in idfobjs1 if getobjname(item) == name]
            n_idfobjs2 = [item for item in idfobjs2 if getobjname(item) == name]
            for idfobj1, idfobj2 in zip_longest(n_idfobjs1, n_idfobjs2):
                if idfobj1 == None:
                    thediffs[(idfobj2.key.upper(), getobjname(idfobj2))] = (
                        None,
                        idf2.idfname,
                    )  # (idf1.idfname, None) -> old
                    break
                if idfobj2 == None:
                    thediffs[(idfobj1.key.upper(), getobjname(idfobj1))] = (
                        idf1.idfname,
                        None,
                    )  # (None, idf2.idfname) -> old
                    break
                for i, (f1, f2) in enumerate(zip(idfobj1.obj, idfobj2.obj)):
                    if i == 0:
                        f1, f2 = f1.upper(), f2.upper()
                    if f1 != f2:
                        thediffs[
                            (akey, getobjname(idfobj1), idfobj1.objidd[i]["field"][0])
                        ] = (f1, f2)
    return thediffs


def makecsv(csvdiffs):
    """retun the csv of the diffs"""
    lines = []
    for row in csvdiffs:
        line = ",".join([str(cell) for cell in row])
        lines.append(line)
    return "\n".join(lines)


def printcsv(csvdiffs):
    """print the csv"""
    for row in csvdiffs:
        print(",".join([str(cell) for cell in row]))


def heading2table(soup, table, row):
    """add heading row to table"""
    tr = Tag(soup, name="tr")
    table.append(tr)
    for attr in row:
        th = Tag(soup, name="th")
        tr.append(th)
        th.append(attr)


def row2table(soup, table, row):
    """ad a row to the table"""
    tr = Tag(soup, name="tr")
    table.append(tr)
    for attr in row:
        td = Tag(soup, name="td")
        tr.append(td)
        td.append(attr)


def makehtmlsoup(csvdiffs):
    """make the html soup"""
    soup = BeautifulSoup()
    html = Tag(soup, name="html")
    para1 = Tag(soup, name="p")
    para1.append(csvdiffs[0][0])
    para2 = Tag(soup, name="p")
    para2.append(csvdiffs[1][0])
    table = Tag(soup, name="table")
    table.attrs.update(dict(border="1"))

    soup.append(html)
    html.append(para1)
    html.append(para2)
    html.append(table)
    heading2table(soup, table, csvdiffs[3])
    for row in csvdiffs[4:]:
        row = [str(cell) for cell in row]
        row2table(soup, table, row)
    # print soup.prettify()
    return soup


def printhtml(csvdiffs):
    """print the html"""
    soup = makehtmlsoup(csvdiffs)
    print(soup)


def htmlinnotebook(soup):
    """display the html in jupyter notebook"""
    from IPython.core.display import display, HTML

    soupstr = str(soup)
    display(HTML(soupstr))


if __name__ == "__main__":
    # do the argparse stuff
    parser = argparse.ArgumentParser(usage=None, description=__doc__)
    parser.add_argument(
        "file1",
        action="store",
        help="location of first with idf files = ./somewhere/f1.idf",
    )
    parser.add_argument(
        "file2",
        action="store",
        help="location of second with idf files = ./somewhere/f2.idf",
    )
    parser.add_argument(
        "--idd",
        action="store",
        help="location of idd file = ./somewhere/eplusv8-0-1.idd",
    )
    # parser.add_argument(
    #     'idd', action='store',
    #     help='location of idd file = ./somewhere/eplusv8-0-1.idd')
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("--csv", action="store_true")
    group.add_argument("--html", action="store_true")
    nspace = parser.parse_args()
    fname1 = nspace.file1
    fname2 = nspace.file2
    iddfile = nspace.idd
    print(iddfile, fname1, fname2)
    # IDF.setiddname(iddfile)
    idf1 = easyopen(fname1, idd=iddfile)
    try:
        idf2 = easyopen(fname2, idd=iddfile)
    except IDDAlreadySetError as e:
        astr = "The two files have different version numners"
        raise IDDMismatchError(astr)

    # TODO What id they have different idd files ?
    thediffs = idfdiffs(idf1, idf2)
    csvdiffs = makecsvdiffs(thediffs, idf1, idf2)
    if nspace.csv:
        printcsv(csvdiffs)
    elif nspace.html:
        printhtml(csvdiffs)


# python idfdiff.py --csv --idd ../resources/iddfiles/Energy+V7_2_0.idd ../resources/idffiles/V_7_2/constr.idf ../resources/idffiles/V_7_2/constr_diff.idf
# python idfdiff.py --html --idd ../resources/iddfiles/Energy+V7_2_0.idd ../resources/idffiles/V_7_2/constr.idf ../resources/idffiles/V_7_2/constr_diff.idf
# python idfdiff2.py --html --idd ../resources/iddfiles/Energy+V8_0_0.idd ../resources/idffiles/V8_0_0/5ZoneSupRetPlenRAB.idf ../resources/idffiles/V8_0_0/5ZoneWaterLoopHeatPump.idf
# python idfdiff.py --csv  ../resources/idffiles/V_7_2/constr.idf ../resources/idffiles/V_7_2/constr_diff.idf
# python idfdiff.py --csv  ../resources/idffiles/V_7_2/constr.idf ../resources/idffiles/V_7_2/constr_diff_bad.idf



# Copyright (c) 2016 Santosh Philip
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
"""integration tests for EPlusInterfaceFunctions"""


from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import eppy.EPlusInterfaceFunctions.iddgroups as iddgroups


def test_idd2group():
    """py.test for idd2group"""
    data = (
        (
            "./eppy/tests/EPlusInterfaceFunctions_tests/integration/iddgroups.idd",
            {
                "G2": ["VersionG", "VersionG1", "VersionG2"],
                "G1": ["Version", "Version1", "Version2"],
                None: ["Lead Input", "Simulation Data"],
            },
        ),  # gdict
    )
    for fname, gdict in data:
        result = iddgroups.idd2group(fname)
        assert result == gdict


def test_idd2grouplist():
    """py.test idd2grouplist"""
    data = (
        (
            "./eppy/tests/EPlusInterfaceFunctions_tests/integration/iddgroups.idd",
            [
                (None, "Lead Input"),
                (None, "Simulation Data"),
                ("G1", "Version"),
                ("G1", "Version1"),
                ("G1", "Version2"),
                ("G2", "VersionG"),
                ("G2", "VersionG1"),
                ("G2", "VersionG2"),
            ],
        ),  # glist
    )
    for fname, glist in data:
        result = iddgroups.idd2grouplist(fname)
        assert result == glist



# Copyright (c) 2016 Jamie Bull
# Copyright (c) 2021 Johan Tibell
# Copyright (c) 2021 Dimitris Mantas
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================
"""Run functions for EnergyPlus.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import platform
import pydoc
import shutil
from subprocess import CalledProcessError, check_call
import sys
import tempfile

from io import StringIO

try:
    import multiprocessing as mp
except ImportError:
    mp = None


def install_paths(version=None, iddname=None):
    """Get the install paths for EnergyPlus executable and weather files.

    We prefer to get the install path from the IDD name but fall back to
    getting it from the version number for backwards compatibility and to
    simplify tests.

    Parameters
    ----------
    version : str, optional
        EnergyPlus version in the format "X-X-X", e.g. "8-7-0".
    iddname : str, optional
        File path to the IDD.

    Returns
    -------
    eplus_exe : str
        Full path to the EnergyPlus executable.
    eplus_weather : str
        Full path to the EnergyPlus weather directory.

    """
    try:
        eplus_exe, eplus_home = paths_from_iddname(iddname)
    except (AttributeError, TypeError, ValueError):
        eplus_exe, eplus_home = paths_from_version(version)
    eplus_weather = os.path.join(eplus_home, "WeatherData")

    return eplus_exe, eplus_weather


def paths_from_iddname(iddname):
    """Get the EnergyPlus install directory and executable path.

    Parameters
    ----------
    iddname : str, optional
        File path to the IDD.

    Returns
    -------
    eplus_exe : str
        Full path to the EnergyPlus executable.
    eplus_home : str
        Full path to the EnergyPlus install directory.

    Raises
    ------
    AttributeError (TypeError on Windows)
        If iddname does not have a directory component (e.g. if None).
    ValueError
        If eplus_exe is not a file.

    """
    eplus_home = os.path.abspath(os.path.dirname(iddname))
    if platform.system() == "Windows":
        eplus_exe = os.path.join(eplus_home, "energyplus.exe")
    elif platform.system() == "Linux":
        eplus_exe = os.path.join(eplus_home, "energyplus")
    else:
        eplus_exe = os.path.join(eplus_home, "energyplus")
    if not os.path.isfile(eplus_exe):
        raise ValueError
    return eplus_exe, eplus_home


def paths_from_version(version):
    """Get the EnergyPlus install directory and executable path.

    Parameters
    ----------
    version : str, optional
        EnergyPlus version in the format "X-X-X", e.g. "8-7-0".

    Returns
    -------
    eplus_exe : str
        Full path to the EnergyPlus executable.
    eplus_home : str
        Full path to the EnergyPlus install directory.

    """
    if platform.system() == "Windows":
        eplus_home = "C:/EnergyPlusV{version}".format(version=version)
        eplus_exe = os.path.join(eplus_home, "energyplus.exe")
    elif platform.system() == "Linux":
        eplus_home = "/usr/local/EnergyPlus-{version}".format(version=version)
        eplus_exe = os.path.join(eplus_home, "energyplus")
    else:
        eplus_home = "/Applications/EnergyPlus-{version}".format(version=version)
        eplus_exe = os.path.join(eplus_home, "energyplus")
    return eplus_exe, eplus_home


def wrapped_help_text(wrapped_func):
    """Decorator to pass through the documentation from a wrapped function."""

    def decorator(wrapper_func):
        """The decorator.

        Parameters
        ----------
        f : callable
            The wrapped function.

        """
        wrapper_func.__doc__ = (
            "This method wraps the following method:\n\n"
            + pydoc.text.document(wrapped_func)
        )
        return wrapper_func

    return decorator


def runIDFs(jobs, processors=1):
    """Wrapper for run() to be used when running IDF5 runs in parallel.

    Parameters
    ----------
    jobs : iterable
        A list or generator made up of an IDF5 object and a kwargs dict
        (see `run_functions.run` for valid keywords).
    processors : int, optional
        Number of processors to run on (default: 1). If 0 is passed then
        the process will run on all CPUs, -1 means one less than all CPUs, etc.

    """
    if mp is None:
        processors = 1
    elif processors <= 0:
        processors = max(1, mp.cpu_count() - processors)

    shutil.rmtree("multi_runs", ignore_errors=True)
    os.mkdir("multi_runs")

    prepared_runs = (
        prepare_run(run_id, run_data) for run_id, run_data in enumerate(jobs)
    )
    if mp is not None:
        pool = mp.Pool(processors)
        if not hasattr(jobs, '__len__'):
            # This avoids materializing all of jobs as a list, potentially
            # use a lot of memory. Since we don't care about the returned
            # results we can use unordered, which is possibly more efficient.
            for _ in pool.imap_unordered(multirunner, prepared_runs):
                pass  # force the whole result to be generated
        else:
            pool.map(multirunner, prepared_runs)
        pool.close()
    else:
        # multiprocessing not present so pass the jobs one at a time
        for job in prepared_runs:
            multirunner([job])
    shutil.rmtree("multi_runs", ignore_errors=True)


def prepare_run(run_id, run_data):
    """Prepare run inputs for one of multiple EnergyPlus runs.

    :param run_id: An ID number for naming the IDF.
    :param run_data: Tuple of the IDF and keyword args to pass to EnergyPlus executable.
    :return: Tuple of the IDF path and EPW, and the keyword args.
    """
    idf, kwargs = run_data
    epw = idf.epw
    idf_dir = os.path.join("multi_runs", "idf_%i" % run_id)
    os.mkdir(idf_dir)
    idf_path = os.path.join(idf_dir, "in.idf")
    idf.saveas(idf_path)
    return (idf_path, epw), kwargs


def multirunner(args):
    """Wrapper for run() to be used when running IDF and EPW runs in parallel.

    Parameters
    ----------
    args : list
        A list made up of a two-item list (IDF and EPW) and a kwargs dict.

    """
    run(*args[0], **args[1])


def run(
    idf=None,
    weather=None,
    output_directory="",
    annual=False,
    design_day=False,
    idd=None,
    epmacro=False,
    expandobjects=False,
    readvars=False,
    output_prefix=None,
    output_suffix=None,
    version=False,
    verbose="v",
    ep_version=None,
):
    """
    Wrapper around the EnergyPlus command line interface.

    Parameters
    ----------
    idf : str
        Full or relative path to the IDF file to be run, or an IDF object.

    weather : str
        Full or relative path to the weather file.

    output_directory : str, optional
        Full or relative path to an output directory (default: 'run_outputs)

    annual : bool, optional
        If True then force annual simulation (default: False)

    design_day : bool, optional
        Force design-day-only simulation (default: False)

    idd : str, optional
        Input data dictionary (default: Energy+.idd in EnergyPlus directory)

    epmacro : str, optional
        Run EPMacro prior to simulation (default: False).

    expandobjects : bool, optional
        Run ExpandObjects prior to simulation (default: False)

    readvars : bool, optional
        Run ReadVarsESO after simulation (default: False)

    output_prefix : str, optional
        Prefix for output file names (default: eplus)

    output_suffix : str, optional
        Suffix style for output file names (default: L)
            L: Legacy (e.g., eplustbl.csv)
            C: Capital (e.g., eplusTable.csv)
            D: Dash (e.g., eplus-table.csv)

    version : bool, optional
        Display version information (default: False)

    verbose: str
        Set verbosity of runtime messages (default: v)
            v: verbose
            q: quiet
            s: silent

    ep_version: str
        EnergyPlus version, used to find install directory. Required if run() is
        called with an IDF file path rather than an IDF object.

    Returns
    -------
    str : status

    Raises
    ------
    CalledProcessError

    AttributeError
        If no ep_version parameter is passed when calling with an IDF file path
        rather than an IDF object.


    """
    args = locals().copy()
    # get unneeded params out of args ready to pass the rest to energyplus.exe
    verbose = args.pop("verbose").lower()
    idf = args.pop("idf")
    iddname = args.get("idd")
    if not isinstance(iddname, str):
        args.pop("idd")
    try:
        idf_path = os.path.abspath(idf.idfname)
    except AttributeError:
        idf_path = os.path.abspath(idf)
    if not os.path.isfile(idf_path):
        raise EnergyPlusRunError(
            "ERROR: Could not find input data file: {}".format(idf_path)
        )
    if not expandobjects:
        with open(idf_path, "r") as f:
            args["expandobjects"] = "HVACTEMPLATE:" in f.read().upper()
    ep_version = args.pop("ep_version")
    # get version from IDF object or by parsing the IDF file for it
    if not ep_version:
        try:
            ep_version = "-".join(str(x) for x in idf.idd_version[:3])
        except AttributeError:
            raise AttributeError(
                "The ep_version must be set when passing an IDF path. \
                Alternatively, use IDF.run()"
            )

    eplus_exe_path, eplus_weather_path = install_paths(ep_version, iddname)
    if version:
        # just get EnergyPlus version number and return
        cmd = [eplus_exe_path, "--version"]
        check_call(cmd)
        return

    # convert paths to absolute paths if required
    if os.path.isfile(args["weather"]):
        args["weather"] = os.path.abspath(args["weather"])
    else:
        args["weather"] = os.path.join(eplus_weather_path, args["weather"])
    output_dir = os.path.abspath(args["output_directory"])
    args["output_directory"] = output_dir
    if iddname is not None:
        args["idd"] = os.path.abspath(iddname)

    # store the directory we start in
    cwd = os.getcwd()
    run_dir = os.path.abspath(tempfile.mkdtemp())
    os.chdir(run_dir)

    # store the output prefix, as it influences the error file name
    output_prefix = args.get("output_prefix")

    # build a list of command line arguments
    cmd = [eplus_exe_path]
    for arg in args:
        if args[arg]:
            if isinstance(args[arg], bool):
                args[arg] = ""
            cmd.extend(["--{}".format(arg.replace("_", "-"))])
            if args[arg] != "":
                cmd.extend([args[arg]])
    cmd.extend([idf_path])

    # send stdout to tmp filehandle to avoid issue #245
    tmp_err = StringIO()
    old_err = sys.stderr
    sys.stderr = tmp_err
    try:
        if verbose == "v":
            print("\r\n" + " ".join(cmd) + "\r\n")
            check_call(cmd)
        elif verbose == "q":
            check_call(cmd, stdout=open(os.devnull, "w"))
        elif verbose == "s":
            with open(os.devnull, "w") as null:
                # Null can be written to, so this is not expected to affect issue #245.
                check_call(cmd, stdout=null, stderr=null)
        else:
            raise ValueError("Unknown verbose mode: {}".format(verbose))
    except CalledProcessError:
        if output_prefix:
            err_file = os.path.join(output_dir, output_prefix + ".err")
        else:
            err_file = os.path.join(output_dir, "eplusout.err")
        message = parse_error(tmp_err, err_file)
        raise EnergyPlusRunError(message)
    finally:
        sys.stderr = old_err
        os.chdir(cwd)
    return "OK"


def parse_error(tmp_err, err_file):
    """Add contents of stderr and eplusout.err and put it in the exception message.

    :param tmp_err: file-like
    :param err_file: str
    :return: str
    """
    std_err = tmp_err.getvalue()
    if os.path.isfile(err_file):
        with open(err_file, "r") as f:
            ep_err = f.read()
    else:
        ep_err = "<File not found>"
    message = "\r\n{std_err}\r\nContents of EnergyPlus error file at {err_file}\r\n{ep_err}".format(
        **locals()
    )
    return message


class EnergyPlusRunError(Exception):
    pass



# Copyright (c) 2012 Santosh Philip
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================

"""py.test for examples. (ex_*.py files)"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from eppy.idfreader import idfreader
import eppy.snippet as snippet
import os
import eppy.pytest_helpers as pytest_helpers
import eppy.modeleditor as modeleditor

# iddfile = "../iddfiles/Energy+V7_0_0_036.idd"
# fname = "../idffiles/V_7_0/5ZoneSupRetPlenRAB.idf"


# iddsnippet = snippet.iddsnippet
from eppy.iddcurrent import iddcurrent

iddsnippet = iddcurrent.iddtxt

idfsnippet = snippet.idfsnippet

from io import StringIO

idffhandle = StringIO(idfsnippet)
iddfhandle = StringIO(iddsnippet)
bunchdt, data, commdct, idd_index = idfreader(idffhandle, iddfhandle)


def test_readwrite():
    """py.test for ex_readwrite"""
    txt = str(data)
    head = "Zone,\n     PLENUM-1,\n     0.0,\n     0.0,\n     0.0,\n     0.0,\n     1,\n     1,\n     0.609600067,\n     283.2;\n\n"
    tail = "\n\nBuildingSurface:Detailed,\n     WALL-1PF,\n     WALL,\n     WALL-1,\n     PLENUM-1,\n     Outdoors,\n     ,\n     SunExposed,\n     WindExposed,\n     0.5,\n     4,\n     0.0,\n     0.0,\n     3.0,\n     0.0,\n     0.0,\n     2.4,\n     30.5,\n     0.0,\n     2.4,\n     30.5,\n     0.0,\n     3.0;\n\n"
    # assert head == txt[:108]
    # assert tail == txt[-280:]


def test_pythonic():
    """py.test for ex_pythonic.py"""
    zones = bunchdt["zone"]  # all the zones
    zone0 = zones[0]
    # -
    printout = "PLENUM-1"
    assert zone0.Name == printout
    # -
    printout = [
        "PLENUM-1",
        "SPACE1-1",
        "SPACE2-1",
        "SPACE3-1",
        "SPACE4-1",
        "SPACE5-1",
        "Sup-PLENUM-1",
    ]
    zonenames = [zone.Name for zone in zones]
    assert printout == zonenames
    # -
    printout = [
        "283.2",
        "239.247360229",
        "103.311355591",
        "239.247360229",
        "103.311355591",
        "447.682556152",
        "208.6",
    ]
    zonevolumes = [zone.Volume for zone in zones]
    for item1, item2 in zip(printout, zonevolumes):
        item1, item2 = float(item1), float(item2)
        assert pytest_helpers.almostequal(item1, item2)
    # -
    printout = [("SPACE2-1", "103.311355591"), ("SPACE4-1", "103.311355591")]
    smallzones = [zn for zn in zones if float(zn.Volume) < 150]
    namevolume = [(zn.Name, zn.Volume) for zn in smallzones]
    for (n1, v1), (n2, v2) in zip(printout, namevolume):
        (n1, v1) = (n1, float(v1))
        (n2, v2) = (n2, float(v2))
        assert n1 == n2
        assert pytest_helpers.almostequal(v1, v2)
    # -
    printout = 2
    assert printout == len(smallzones)
    # -
    printout = [
        "PLENUM-1",
        "SPACE1-1",
        "FIRST-SMALL-ZONE",
        "SPACE3-1",
        "SECOND-SMALL-ZONE",
        "SPACE5-1",
        "Sup-PLENUM-1",
    ]
    smallzones[0].Name = "FIRST-SMALL-ZONE"
    smallzones[1].Name = "SECOND-SMALL-ZONE"
    # now the zone names are:
    zonenames = [zone.Name for zone in zones]
    assert printout == zonenames


def test_addobject():
    """py.test for ex_addobject.py"""
    zones = bunchdt["zone"]  # all the zones
    assert len(zones) == 7
    modeleditor.addobject(bunchdt, data, commdct, "Zone".upper(), None, aname="NewZone")
    assert len(zones) == 8
    assert zones[-1].obj == [
        "ZONE",
        "NewZone",
        0.0,
        0.0,
        0.0,
        0.0,
        1,
        1,
        "autocalculate",
        "autocalculate",
        "autocalculate",
        "",
        "",
        "Yes",
    ]


def test_functions():
    """py.test for ex_functions.py"""
    surfaces = bunchdt["BuildingSurface:Detailed"]  # all the surfaces
    assert len(surfaces) == 1
    surface = surfaces[0]
    assert surface.Name == "WALL-1PF"
    assert surface.azimuth == 180.0
    assert surface.tilt == 90.0
    assert pytest_helpers.almostequal(surface.area, 18.3)



"""find all related objects - skiping extensible fields
this script is to assist in further coding - not for end user
Will assist in:
    - discovering useful functions to add to EpBunch using addfunctions
"""
# was designed to run from ../../docs
# to run from here change some of the file paths below
# run as
# python relatedbojects.py > outputfile.txt
# look at outputfile.txt
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import sys

# pathnameto_eppy = 'c:/eppy'
pathnameto_eppy = "../"
sys.path.append(pathnameto_eppy)

from eppy import modeleditor
from eppy.modeleditor import IDF
import StringIO

iddfile = "./eppy/resources/iddfiles/Energy+V7_2_0.idd"
iddfile = "./eppy/resources/iddfiles/Energy+V8_0_0.idd"

IDF.setiddname(iddfile)
idf = IDF(StringIO.StringIO(""))

okeys = idf.idfobjects.keys()
for okey in okeys:
    idf.newidfobject(okey)

for okey in okeys:
    referreds = idf.idfobjects[okey.upper()]
    referred = referreds[0]
    try:
        nameidd = referred.getfieldidd("Name")
    except ValueError as e:
        # print 'no Name', okey
        continue
    try:
        references = nameidd["reference"]
    except KeyError as e:
        # print 'no reference', okey
        continue

    groupfield = {}
    ikeys = idf.idfobjects.keys()
    for key in ikeys:
        # key = u'BUILDINGSURFACE:DETAILED'
        # key = u'ZONELIST'
        objs = idf.idfobjects[key]
        obj = objs[0]
        keyidd = obj.getfieldidd("key")
        group = keyidd["group"]

        fields = obj.objls
        for field in fields:
            # field = u'Zone_1_Name'
            fieldidd = obj.getfieldidd(field)
            try:
                referrings = fieldidd["object-list"]
            except KeyError as e:
                referrings = []
            if fieldidd.has_key("begin-extensible"):
                # print field, fieldidd['begin-extensible']
                break  # do extensibles later
            s_references = set(references)
            s_referrings = set(referrings)

            intersect = s_references.intersection(s_referrings)
            if intersect:
                akey = (group, field)
                if not groupfield.has_key(akey):
                    groupfield[akey] = []
                    groupfield[akey].append(key)
                else:
                    groupfield[akey].append(key)
                # print obj.key
                # print '\t', dict(group=group, field=field)
                # print

    for kkey, values in groupfield.items():
        print(okey)
        print("\t", "->", kkey)
        for value in values:
            print("\t\t", value)
        print



# Copyright (c) 2016 Santosh Philip
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================
"""py.test for idd_index"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals


from eppy.EPlusInterfaceFunctions import iddindex

commdct = [
    [
        {
            "format": ["singleLine"],
            "group": "Simulation Parameters",
            "idfobj": "Version",
            "memo": ["Specifies the EnergyPlus version of the IDF file."],
            "unique-object": [""],
        },
        {"default": ["7.0"], "field": ["Version Identifier"], "required-field": [""]},
    ],
    [
        {
            "group": "Simulation Parameters",
            "idfobj": "Building",
            "memo": [
                "Describes parameters that are used during the simulation",
                "of the building. There are necessary correlations between the entries for",
                "this object and some entries in the Site:WeatherStation and",
                "Site:HeightVariation objects, specifically the Terrain field.",
            ],
            "min-fields": ["8"],
            "required-object": [""],
            "unique-object": [""],
        },
        {
            "default": ["NONE"],
            "field": ["Name"],
            "required-field": [""],
            "retaincase": [""],
        },
        {
            "default": ["0.0"],
            "field": ["North Axis"],
            "note": ["degrees from true North"],
            "type": ["real"],
            "units": ["deg"],
        },
    ],
    [
        {
            "format": ["vertices"],
            "group": "Thermal Zones and Surfaces",
            "idfobj": "Zone",
            "memo": ["Defines a thermal zone of the building."],
        },
        {
            "field": ["Name"],
            "reference": [
                "ZoneNames",
                "OutFaceEnvNames",
                "ZoneAndZoneListNames",
                "AirflowNetworkNodeAndZoneNames",
            ],
            "required-field": [""],
            "type": ["alpha"],
        },
        {
            "default": ["0"],
            "field": ["Direction of Relative North"],
            "type": ["real"],
            "units": ["deg"],
        },
    ],
    [
        {
            "extensible:3": [
                '-- duplicate last set of x,y,z coordinates (last 3 fields), remembering to remove ; from "inner" fields.'
            ],
            "format": ["vertices"],
            "group": "Thermal Zones and Surfaces",
            "idfobj": "BuildingSurface:Detailed",
            "memo": [
                "Allows for detailed entry of building heat transfer surfaces. Does not include subsurfaces such as windows or doors."
            ],
            "min-fields": ["19"],
        },
        {
            "field": ["Name"],
            "reference": [
                "SurfaceNames",
                "SurfAndSubSurfNames",
                "AllHeatTranSurfNames",
                "HeatTranBaseSurfNames",
                "OutFaceEnvNames",
                "AllHeatTranAngFacNames",
                "RadGroupAndSurfNames",
                "SurfGroupAndHTSurfNames",
                "AllShadingAndHTSurfNames",
            ],
            "required-field": [""],
            "type": ["alpha"],
        },
        {
            "field": ["Surface Type"],
            "key": ["Floor", "Wall", "Ceiling", "Roof"],
            "required-field": [""],
            "type": ["choice"],
        },
    ],
    [
        {
            "format": ["vertices"],
            "group": "Thermal Zones and Surfaces",
            "idfobj": "FenestrationSurface:Detailed",
            "memo": [
                "Allows for detailed entry of subsurfaces",
                "(windows, doors, glass doors, tubular daylighting devices).",
            ],
            "min-fields": ["19"],
        },
        {
            "field": ["Name"],
            "reference": [
                "SubSurfNames",
                "SurfAndSubSurfNames",
                "AllHeatTranSurfNames",
                "OutFaceEnvNames",
                "AllHeatTranAngFacNames",
                "RadGroupAndSurfNames",
                "SurfGroupAndHTSurfNames",
                "AllShadingAndHTSurfNames",
            ],
            "required-field": [""],
            "type": ["alpha"],
        },
        {
            "field": ["Surface Type"],
            "key": [
                "Window",
                "Door",
                "GlassDoor",
                "TubularDaylightDome",
                "TubularDaylightDiffuser",
            ],
            "required-field": [""],
            "type": ["choice"],
        },
    ],
    [
        {
            "group": "Thermal Zones and Surfaces",
            "idfobj": "Wall:Exterior",
            "memo": [
                "Allows for simplified entry of exterior walls.",
                "View Factor to Ground is automatically calculated.",
            ],
        },
        {
            "field": ["Name"],
            "reference": [
                "SurfaceNames",
                "SurfAndSubSurfNames",
                "AllHeatTranSurfNames",
                "HeatTranBaseSurfNames",
                "AllHeatTranAngFacNames",
                "RadGroupAndSurfNames",
                "SurfGroupAndHTSurfNames",
                "AllShadingAndHTSurfNames",
            ],
            "required-field": [""],
            "type": ["alpha"],
        },
        {
            "field": ["Construction Name"],
            "note": ["To be matched with a construction in this input file"],
            "object-list": ["ConstructionNames"],
            "required-field": [""],
            "type": ["object-list"],
        },
    ],
    [
        {
            "group": "Thermal Zones and Surfaces",
            "idfobj": "Window",
            "memo": ["Allows for simplified entry of Windows."],
        },
        {
            "field": ["Name"],
            "reference": [
                "SubSurfNames",
                "SurfAndSubSurfNames",
                "AllHeatTranSurfNames",
                "OutFaceEnvNames",
                "AllHeatTranAngFacNames",
                "RadGroupAndSurfNames",
                "SurfGroupAndHTSurfNames",
                "AllShadingAndHTSurfNames",
            ],
            "required-field": [""],
            "type": ["alpha"],
        },
        {
            "field": ["Construction Name"],
            "note": ["To be matched with a construction in this input file"],
            "object-list": ["ConstructionNames"],
            "required-field": [""],
            "type": ["object-list"],
        },
    ],
]


def test_makename2refdct():
    """py.test for makename2refdct"""
    # do a simple test
    thedata = (
        (
            {"ZONE": ["Z1", "Z2"]},
            [[{"idfobj": "zone"}, {"field": ["Name"], "reference": ["Z1", "Z2"]}]],
        ),  # expected, simpledct
        (
            {"ZONE": ["Z1", "Z2"], "WALL": ["W1", "W2"]},
            [
                [{"idfobj": "zone"}, {"field": ["Name"], "reference": ["Z1", "Z2"]}],
                [{"idfobj": "wall"}, {"field": ["Name"], "reference": ["W1", "W2"]}],
            ],
        ),  # expected, simpledct
        (
            {"ZONE": ["Z1", "Z2"], "WALL": ["W1", "W2"]},
            [
                [{"idfobj": "zone"}, {"field": ["Name"], "reference": ["Z1", "Z2"]}],
                [{"idfobj": "wall"}, {"field": ["Name"], "reference": ["W1", "W2"]}],
                [],  # put in random stuff
            ],
        ),  # expected, simpledct
        (
            {"WALL": ["W1", "W2"]},
            [
                [{"idfobj": "zone"}, {"field": ["notName"], "reference": ["Z1", "Z2"]}],
                [{"idfobj": "wall"}, {"field": ["Name"], "reference": ["W1", "W2"]}],
                [],  # put in random stuff
            ],
        ),  # expected, simpledct
        (
            {},
            [
                [{"idfobj": "zone"}, {"field": ["notName"], "reference": ["Z1", "Z2"]}],
                [{"idfobj": "wall"}, {"field": ["Name"], "noreference": ["W1", "W2"]}],
                [],  # put in random stuff
            ],
        ),  # expected, simpledct
    )
    for expected, simpledct in thedata:
        result = iddindex.makename2refdct(simpledct)
        assert result == expected
    # the test with real data
    expected = {
        "ZONE": [
            "ZoneNames",
            "OutFaceEnvNames",
            "ZoneAndZoneListNames",
            "AirflowNetworkNodeAndZoneNames",
        ],
        "WINDOW": [
            "SubSurfNames",
            "SurfAndSubSurfNames",
            "AllHeatTranSurfNames",
            "OutFaceEnvNames",
            "AllHeatTranAngFacNames",
            "RadGroupAndSurfNames",
            "SurfGroupAndHTSurfNames",
            "AllShadingAndHTSurfNames",
        ],
        "WALL:EXTERIOR": [
            "SurfaceNames",
            "SurfAndSubSurfNames",
            "AllHeatTranSurfNames",
            "HeatTranBaseSurfNames",
            "AllHeatTranAngFacNames",
            "RadGroupAndSurfNames",
            "SurfGroupAndHTSurfNames",
            "AllShadingAndHTSurfNames",
        ],
        "FENESTRATIONSURFACE:DETAILED": [
            "SubSurfNames",
            "SurfAndSubSurfNames",
            "AllHeatTranSurfNames",
            "OutFaceEnvNames",
            "AllHeatTranAngFacNames",
            "RadGroupAndSurfNames",
            "SurfGroupAndHTSurfNames",
            "AllShadingAndHTSurfNames",
        ],
        "BUILDINGSURFACE:DETAILED": [
            "SurfaceNames",
            "SurfAndSubSurfNames",
            "AllHeatTranSurfNames",
            "HeatTranBaseSurfNames",
            "OutFaceEnvNames",
            "AllHeatTranAngFacNames",
            "RadGroupAndSurfNames",
            "SurfGroupAndHTSurfNames",
            "AllShadingAndHTSurfNames",
        ],
    }
    result = iddindex.makename2refdct(commdct)
    assert result == expected


def test_makeref2namesdct():
    """pytest for makeref2namesdct"""
    thedata = (
        (
            {
                "wall": ["surface", "surfandsubsurf"],
                "roof": ["surface", "surfandsubsurf"],
                "window": ["surfandsubsurf", "subsurf"],
                "skylight": ["surfandsubsurf", "subsurf"],
                "zone": ["zname"],
            },
            {
                "surface": set(["wall", "roof"]),
                "subsurf": set(["window", "skylight"]),
                "surfandsubsurf": set(["wall", "roof", "window", "skylight"]),
                "zname": set(["zone"]),
            },
        ),  # name2refdct, expected
        (
            {
                "ZONE": [
                    "ZoneNames",
                    "OutFaceEnvNames",
                    "ZoneAndZoneListNames",
                    "AirflowNetworkNodeAndZoneNames",
                ],
                "WINDOW": [
                    "SubSurfNames",
                    "SurfAndSubSurfNames",
                    "AllHeatTranSurfNames",
                    "OutFaceEnvNames",
                    "AllHeatTranAngFacNames",
                    "RadGroupAndSurfNames",
                    "SurfGroupAndHTSurfNames",
                    "AllShadingAndHTSurfNames",
                ],
                "WALL:EXTERIOR": [
                    "SurfaceNames",
                    "SurfAndSubSurfNames",
                    "AllHeatTranSurfNames",
                    "HeatTranBaseSurfNames",
                    "AllHeatTranAngFacNames",
                    "RadGroupAndSurfNames",
                    "SurfGroupAndHTSurfNames",
                    "AllShadingAndHTSurfNames",
                ],
                "FENESTRATIONSURFACE:DETAILED": [
                    "SubSurfNames",
                    "SurfAndSubSurfNames",
                    "AllHeatTranSurfNames",
                    "OutFaceEnvNames",
                    "AllHeatTranAngFacNames",
                    "RadGroupAndSurfNames",
                    "SurfGroupAndHTSurfNames",
                    "AllShadingAndHTSurfNames",
                ],
                "BUILDINGSURFACE:DETAILED": [
                    "SurfaceNames",
                    "SurfAndSubSurfNames",
                    "AllHeatTranSurfNames",
                    "HeatTranBaseSurfNames",
                    "OutFaceEnvNames",
                    "AllHeatTranAngFacNames",
                    "RadGroupAndSurfNames",
                    "SurfGroupAndHTSurfNames",
                    "AllShadingAndHTSurfNames",
                ],
            },
            {
                "AirflowNetworkNodeAndZoneNames": {"ZONE"},
                "AllHeatTranAngFacNames": {
                    "BUILDINGSURFACE:DETAILED",
                    "FENESTRATIONSURFACE:DETAILED",
                    "WALL:EXTERIOR",
                    "WINDOW",
                },
                "AllHeatTranSurfNames": {
                    "BUILDINGSURFACE:DETAILED",
                    "FENESTRATIONSURFACE:DETAILED",
                    "WALL:EXTERIOR",
                    "WINDOW",
                },
                "AllShadingAndHTSurfNames": {
                    "BUILDINGSURFACE:DETAILED",
                    "FENESTRATIONSURFACE:DETAILED",
                    "WALL:EXTERIOR",
                    "WINDOW",
                },
                "HeatTranBaseSurfNames": {"BUILDINGSURFACE:DETAILED", "WALL:EXTERIOR"},
                "OutFaceEnvNames": {
                    "BUILDINGSURFACE:DETAILED",
                    "FENESTRATIONSURFACE:DETAILED",
                    "WINDOW",
                    "ZONE",
                },
                "RadGroupAndSurfNames": {
                    "BUILDINGSURFACE:DETAILED",
                    "FENESTRATIONSURFACE:DETAILED",
                    "WALL:EXTERIOR",
                    "WINDOW",
                },
                "SubSurfNames": {"FENESTRATIONSURFACE:DETAILED", "WINDOW"},
                "SurfAndSubSurfNames": {
                    "BUILDINGSURFACE:DETAILED",
                    "FENESTRATIONSURFACE:DETAILED",
                    "WALL:EXTERIOR",
                    "WINDOW",
                },
                "SurfGroupAndHTSurfNames": {
                    "BUILDINGSURFACE:DETAILED",
                    "FENESTRATIONSURFACE:DETAILED",
                    "WALL:EXTERIOR",
                    "WINDOW",
                },
                "SurfaceNames": {"BUILDINGSURFACE:DETAILED", "WALL:EXTERIOR"},
                "ZoneAndZoneListNames": {"ZONE"},
                "ZoneNames": {"ZONE"},
            },
        ),  # name2refdct, expected
    )
    for name2refdct, expected in thedata:
        result = iddindex.makeref2namesdct(name2refdct)
        assert result == expected


def test_ref2names2commdct():
    """py.test for ref2names2commdct"""
    thedata = (
        (
            # ------------
            [
                [
                    {"idfobj": "referedto1"},
                    {
                        "field": ["Name"],
                        "reference": ["rname11", "rname12", "rname_both"],
                    },
                ],
                [
                    {"idfobj": "referedto2"},
                    {
                        "field": ["Name"],
                        "reference": ["rname21", "rname22", "rname_both"],
                    },
                ],
                [
                    {"idfobj": "referingobj1"},
                    {"field": ["Name"]},
                    {
                        "field": ["referingfield"],
                        "type": ["object-list"],
                        "object-list": ["rname11"],
                    },
                ],
                [
                    {"idfobj": "referingobj2"},
                    {"field": ["Name"]},
                    {
                        "field": ["referingfield"],
                        "type": ["object-list"],
                        "object-list": ["rname_both"],
                    },
                ],
            ],
            # ------------
            [
                [
                    {"idfobj": "referedto1"},
                    {
                        "field": ["Name"],
                        "reference": ["rname11", "rname12", "rname_both"],
                    },
                ],
                [
                    {"idfobj": "referedto2"},
                    {
                        "field": ["Name"],
                        "reference": ["rname21", "rname22", "rname_both"],
                    },
                ],
                [
                    {"idfobj": "referingobj1"},
                    {"field": ["Name"]},
                    {
                        "field": ["referingfield"],
                        "type": ["object-list"],
                        "object-list": ["rname11"],
                        "validobjects": set(["referedto1".upper()]),
                    },
                ],
                [
                    {"idfobj": "referingobj2"},
                    {"field": ["Name"]},
                    {
                        "field": ["referingfield"],
                        "type": ["object-list"],
                        "object-list": ["rname_both"],
                        "validobjects": set(["REFEREDTO1", "REFEREDTO2"]),
                    },
                ],
            ],
        ),  # commdct, expected
    )
    for commdct, expected in thedata:
        name2refdct = iddindex.makename2refdct(commdct)
        ref2names = iddindex.makeref2namesdct(name2refdct)
        result = iddindex.ref2names2commdct(ref2names, commdct)
        for r_item, e_item in zip(result, expected):
            assert r_item == e_item
            # the test below is ensure that the embedded data is not a copy,
            # but is pointing to the set in ref2names
            for item in r_item:
                try:
                    reference = item["object-list"][0]
                    validobjects = item["validobjects"]
                    assert id(ref2names[reference]) == id(validobjects)
                except KeyError as e:
                    continue



# Copyright (c) 2019 Santosh Philip
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================
"""py.test for idd_helpers"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals


import eppy.idd_helpers as idd_helpers


def test_folder2ver():
    """py.test for folder2ver"""
    data = (
        ("EnergyPlus-8-8-0", "8.8.0"),  # folder, expected
        ("EnergyPlusV8-8-0", "8.8.0"),  # folder, expected
    )
    for folder, expected in data:
        result = idd_helpers.folder2ver(folder)
        assert result == expected



# Copyright (c) 2012 Santosh Philip
# Copyright (c) 2021 Jeremy Lerond
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================
"""py.test for modeleditor"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from itertools import product
import os
import warnings
import os.path
import shutil
from pathlib import Path


import pytest
from six import StringIO
from six import string_types

from eppy import modeleditor
from eppy.bunch_subclass import Bunch
from eppy.iddcurrent import iddcurrent
from eppy.modeleditor import IDF
from eppy.pytest_helpers import almostequal
import eppy.snippet as snippet


iddsnippet = iddcurrent.iddtxt
idfsnippet = snippet.idfsnippet

# idffhandle = StringIO(idfsnippet)
# iddfhandle = StringIO(iddsnippet)
# bunchdt, data, commdct, gdict = idfreader.idfreader(idffhandle, iddfhandle, None)

# idd is read only once in this test
# if it has already been read from some other test, it will continue with
# the old reading
iddfhandle = StringIO(iddcurrent.iddtxt)
if IDF.getiddname() == None:
    IDF.setiddname(iddfhandle)


def test_poptrailing():
    """py.test for poptrailing"""
    tdata = (
        ([1, 2, 3, "", 56, "", "", "", ""], [1, 2, 3, "", 56]),  # lst, popped
        ([1, 2, 3, "", 56], [1, 2, 3, "", 56]),  # lst, popped
        ([1, 2, 3, 56], [1, 2, 3, 56]),  # lst, popped
    )
    for before, after in iter(tdata):
        assert modeleditor.poptrailing(before) == after


def test_extendlist():
    """py.test for extendlist"""
    tdata = (
        ([1, 2, 3], 2, 0, [1, 2, 3]),  # lst, i, value, nlst
        ([1, 2, 3], 3, 0, [1, 2, 3, 0]),  # lst, i, value, nlst
        ([1, 2, 3], 5, 0, [1, 2, 3, 0, 0, 0]),  # lst, i, value, nlst
        ([1, 2, 3], 7, 0, [1, 2, 3, 0, 0, 0, 0, 0]),  # lst, i, value, nlst
    )
    for lst, i, value, nlst in tdata:
        modeleditor.extendlist(lst, i, value=value)
        assert lst == nlst


def test_namebunch():
    """py.test for namebunch"""
    thedata = (
        (Bunch(dict(Name="", a=5)), "yay", "yay"),  # abunch, aname, thename
        (Bunch(dict(Name=None, a=5)), "yay", None),  # abunch, aname, thename
    )
    for abunch, aname, thename in thedata:
        result = modeleditor.namebunch(abunch, aname)
        assert result.Name == thename


def test_getnamedargs():
    """py.test for getnamedargs"""
    result = dict(a=1, b=2, c=3)
    assert result == modeleditor.getnamedargs(a=1, b=2, c=3)
    assert result == modeleditor.getnamedargs(dict(a=1, b=2, c=3))
    assert result == modeleditor.getnamedargs(dict(a=1, b=2), c=3)
    assert result == modeleditor.getnamedargs(dict(a=1), c=3, b=2)


def test_getrefnames():
    """py.test for getrefnames"""
    tdata = (
        (
            "ZONE",
            [
                "ZoneNames",
                "OutFaceEnvNames",
                "ZoneAndZoneListNames",
                "AirflowNetworkNodeAndZoneNames",
            ],
        ),  # objkey, therefs
        (
            "FluidProperties:Name".upper(),
            ["FluidNames", "FluidAndGlycolNames"],
        ),  # objkey, therefs
        ("Building".upper(), []),  # objkey, therefs
    )
    for objkey, therefs in tdata:
        fhandle = StringIO("")
        idf = IDF(fhandle)
        result = modeleditor.getrefnames(idf, objkey)
        assert result == therefs


def test_getallobjlists():
    """py.test for getallobjlists"""
    tdata = (
        (
            "TransformerNames",
            [("ElectricLoadCenter:Distribution".upper(), "TransformerNames", [10])],
        ),  # refname, objlists
    )
    for refname, objlists in tdata:
        fhandle = StringIO("")
        idf = IDF(fhandle)
        result = modeleditor.getallobjlists(idf, refname)
        assert result == objlists


def test_rename():
    """py.test for rename"""
    idftxt = """Material,
      G01a 19mm gypsum board,  !- Name
      MediumSmooth,            !- Roughness
      0.019,                   !- Thickness {m}
      0.16,                    !- Conductivity {W/m-K}
      800,                     !- Density {kg/m3}
      1090;                    !- Specific Heat {J/kg-K}

      Construction,
        Interior Wall,           !- Name
        G01a 19mm gypsum board,  !- Outside Layer
        F04 Wall air space resistance,  !- Layer 2
        G01a 19mm gypsum board;  !- Layer 3

      Construction,
        Other Wall,           !- Name
        G01a 19mm gypsum board,  !- Outside Layer
        G01a 19mm gypsum board,  !- Layer 2
        G01a 19mm gypsum board;  !- Layer 3

    """
    ridftxt = """Material,
      peanut butter,  !- Name
      MediumSmooth,            !- Roughness
      0.019,                   !- Thickness {m}
      0.16,                    !- Conductivity {W/m-K}
      800,                     !- Density {kg/m3}
      1090;                    !- Specific Heat {J/kg-K}

      Construction,
        Interior Wall,           !- Name
        peanut butter,  !- Outside Layer
        F04 Wall air space resistance,  !- Layer 2
        peanut butter;  !- Layer 3

      Construction,
        Other Wall,           !- Name
        peanut butter,  !- Outside Layer
        peanut butter,  !- Layer 2
        peanut butter;  !- Layer 3

    """
    fhandle = StringIO(idftxt)
    idf = IDF(fhandle)
    result = modeleditor.rename(
        idf, "Material".upper(), "G01a 19mm gypsum board", "peanut butter"
    )
    assert result.Name == "peanut butter"
    assert idf.idfobjects["CONSTRUCTION"][0].Outside_Layer == "peanut butter"
    assert idf.idfobjects["CONSTRUCTION"][0].Layer_3 == "peanut butter"


def test_zonearea_zonevolume():
    """py.test for zonearea and zonevolume"""
    idftxt = """Zone, 473222, 0.0, 0.0, 0.0, 0.0, , 1;
        BuildingSurface:Detailed, F7289B, Floor, Exterior Floor, 473222,
        Ground, ,
        NoSun, NoWind, , 4, 2.23, 2.56, 0.0, 2.23, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
        2.56, 0.0;  BuildingSurface:Detailed, F3659B, Wall, Exterior Wall,
        473222, Outdoors, , SunExposed, WindExposed, , 4, 2.23, 2.56, 1.49,
        2.23, 2.56, 0.0, 0.0, 2.56, 0.0, 0.0, 2.56, 1.49;
        BuildingSurface:Detailed, 46C6C9, Wall, Exterior Wall, 473222,
        Outdoors, , SunExposed, WindExposed, , 4, 2.23, 0.0, 1.49, 2.23,
        0.0, 0.0, 2.23, 1.02548139464, 0.0, 2.23, 1.02548139464, 1.49;
        BuildingSurface:Detailed, 4287DD, Wall, Exterior Wall, 473222,
        Outdoors, , SunExposed, WindExposed, , 4, 0.0, 2.56, 1.49, 0.0,
        2.56, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.49;
        BuildingSurface:Detailed, 570C2E, Wall, Exterior Wall, 473222,
        Outdoors, , SunExposed, WindExposed, , 4, 0.0, 0.0, 1.49, 0.0, 0.0,
        0.0, 2.23, 0.0, 0.0, 2.23, 0.0, 1.49;  BuildingSurface:Detailed,
        BAEA99, Roof, Exterior Roof, 473222, Outdoors, , SunExposed,
        WindExposed, , 4, 0.0, 2.56, 1.49, 0.0, 0.0, 1.49, 2.23, 0.0, 1.49,
        2.23, 2.56, 1.49;  BuildingSurface:Detailed, C879FE, Floor,
        Exterior Floor, 473222, Ground, , NoSun, NoWind, , 4, 3.22,
        2.52548139464, 0.0, 3.22, 1.02548139464, 0.0, 2.23,
        1.02548139464, 0.0, 2.23, 2.52548139464, 0.0;
        BuildingSurface:Detailed, 25B601, Wall, Exterior Wall, 473222,
        Outdoors, , SunExposed, WindExposed, , 4, 2.23,
        1.02548139464, 1.49, 2.23, 1.02548139464, 0.0, 2.23, 2.52548139464,
        0.0, 2.23, 2.52548139464, 1.49;  BuildingSurface:Detailed, F5EADC,
        Wall, Exterior Wall, 473222, Outdoors, , SunExposed, WindExposed, ,
        4, 2.23, 1.02548139464, 1.49, 2.23, 1.02548139464, 0.0, 3.22,
        1.02548139464, 0.0, 3.22, 1.02548139464, 1.49;
        BuildingSurface:Detailed, D0AABE, Wall, Exterior Wall, 473222,
        Outdoors, , SunExposed, WindExposed, , 4, 3.22, 1.02548139464,
        1.49, 3.22, 1.02548139464, 0.0, 3.22, 2.52548139464, 0.0, 3.22,
        2.52548139464, 1.49;  BuildingSurface:Detailed, B0EA02, Wall,
        Exterior Wall, 473222, Outdoors, , SunExposed, WindExposed, ,
        4, 3.22, 2.52548139464, 1.49, 3.22, 2.52548139464, 0.0, 2.23,
        2.52548139464, 0.0, 2.23, 2.52548139464, 1.49;
        BuildingSurface:Detailed, E6DF3B, Roof, Exterior Roof, 473222,
        Outdoors, , SunExposed, WindExposed, , 4, 2.23, 2.52548139464, 1.49,
        2.23, 1.02548139464, 1.49, 3.22, 1.02548139464, 1.49, 3.22,
        2.52548139464, 1.49;  BuildingSurface:Detailed, 4F8681, Wall,
        Exterior Wall, 473222, Outdoors, , SunExposed, WindExposed, , 4,
        2.23, 2.52548139464, 1.49, 2.23, 2.52548139464, 0.0, 2.23, 2.56,
        0.0, 2.23, 2.56, 1.49;  """
    idf = IDF(StringIO(idftxt))
    result = modeleditor.zonearea(idf, "473222")
    assert almostequal(result, 7.1938)
    result = modeleditor.zonearea_floor(idf, "473222")
    assert almostequal(result, 7.1938)
    result = modeleditor.zonearea_roofceiling(idf, "473222")
    assert almostequal(result, 7.1938)
    result = modeleditor.zone_floor2roofheight(idf, "473222")
    assert almostequal(result, 1.49)
    result = modeleditor.zoneheight(idf, "473222")
    assert almostequal(result, 1.49)
    result = modeleditor.zone_floor2roofheight(idf, "473222")
    assert almostequal(result, 1.49)
    result = modeleditor.zonevolume(idf, "473222")
    assert almostequal(result, 10.718762)
    # remove floor
    zone = idf.getobject("ZONE", "473222")
    surfs = idf.idfobjects["BuildingSurface:Detailed".upper()]
    zone_surfs = [s for s in surfs if s.Zone_Name == zone.Name]
    floors = [s for s in zone_surfs if s.Surface_Type.upper() == "FLOOR"]
    for floor in floors:
        idf.removeidfobject(floor)
    result = modeleditor.zonearea_floor(idf, "473222")
    assert almostequal(result, 0)
    result = modeleditor.zonearea_roofceiling(idf, "473222")
    assert almostequal(result, 7.1938)
    result = modeleditor.zonearea(idf, "473222")
    assert almostequal(result, 7.1938)
    result = modeleditor.zoneheight(idf, "473222")
    assert almostequal(result, 1.49)
    result = modeleditor.zonevolume(idf, "473222")
    assert almostequal(result, 10.718762)
    # reload idf and remove roof/ceiling
    idf = IDF(StringIO(idftxt))
    zone = idf.getobject("ZONE", "473222")
    surfs = idf.idfobjects["BuildingSurface:Detailed".upper()]
    zone_surfs = [s for s in surfs if s.Zone_Name == zone.Name]
    roofs = [s for s in zone_surfs if s.Surface_Type.upper() == "ROOF"]
    ceilings = [s for s in zone_surfs if s.Surface_Type.upper() == "CEILING"]
    topsurfaces = roofs + ceilings
    for surf in topsurfaces:
        idf.removeidfobject(surf)
    result = modeleditor.zonearea_roofceiling(idf, "473222")
    assert almostequal(result, 0)
    result = modeleditor.zonearea(idf, "473222")
    assert almostequal(result, 7.1938)
    result = modeleditor.zoneheight(idf, "473222")
    assert almostequal(result, 1.49)
    result = modeleditor.zonevolume(idf, "473222")
    assert almostequal(result, 10.718762)


def test_new():
    """py.test for IDF.new()"""
    idf = IDF()
    idf.new()
    # assert idf.idfobjects['building'.upper()] == Idf_MSequence()
    assert idf.idfobjects["building".upper()].list1 == []
    assert idf.idfobjects["building".upper()].list2 == []


def test_newidfobject():
    """py.test for newidfobject"""
    # make a blank idf
    # make a function for this and then continue.
    idf = IDF()
    idf.new()
    objtype = "material:airgap".upper()
    obj = idf.newidfobject(objtype, Name="Argon")
    obj = idf.newidfobject(objtype, Name="Krypton")
    obj = idf.newidfobject(objtype, Name="Xenon")
    assert idf.model.dt[objtype] == [
        ["MATERIAL:AIRGAP", "Argon"],
        ["MATERIAL:AIRGAP", "Krypton"],
        ["MATERIAL:AIRGAP", "Xenon"],
    ]
    # remove an object
    idf.popidfobject(objtype, 1)
    assert idf.model.dt[objtype] == [
        ["MATERIAL:AIRGAP", "Argon"],
        ["MATERIAL:AIRGAP", "Xenon"],
    ]
    lastobject = idf.idfobjects[objtype][-1]
    idf.removeidfobject(lastobject)
    assert idf.model.dt[objtype] == [["MATERIAL:AIRGAP", "Argon"]]
    # copyidfobject
    onlyobject = idf.idfobjects[objtype][0]
    idf.copyidfobject(onlyobject)

    assert idf.model.dt[objtype] == [
        ["MATERIAL:AIRGAP", "Argon"],
        ["MATERIAL:AIRGAP", "Argon"],
    ]
    # remove all objects
    idf.removeallidfobjects(objtype)
    assert len(idf.idfobjects[objtype]) == 0
    # test some functions
    objtype = "FENESTRATIONSURFACE:DETAILED"
    obj = idf.newidfobject(objtype, Name="A Wall")
    assert obj.coords == []
    assert obj.fieldvalues[1] == "A Wall"

    # test defaultvalues=True and defaultvalues=False
    sim_deftrue = idf.newidfobject("SimulationControl".upper(), defaultvalues=True)
    assert sim_deftrue.Do_Zone_Sizing_Calculation == "No"
    sim_deffalse = idf.newidfobject("SimulationControl".upper(), defaultvalues=False)
    assert sim_deffalse.Do_Zone_Sizing_Calculation == ""


def test_newidfobject_warning():
    """Test that the warning for newidfobject created with `aname` is working.

    Fails if the warning is not issued when `aname` is used, or if the warning
    is issued when `aname` is not used.
    """
    # make a blank idf
    # make a function for this and then continue.
    idf = IDF()
    idf.new()
    objtype = "material:airgap".upper()
    # expect warnings here
    with pytest.warns(UserWarning):
        idf.newidfobject(objtype, aname="Krypton")
    with pytest.warns(UserWarning):
        idf.newidfobject(objtype, "Krypton")

    # expect no warnings here - we pass None so as not to trigger the `Failed: DID NOT WARN` message from pytest
    with pytest.warns(None) as captured_warnings:
        idf.newidfobject(objtype, Name="Krypton")
    assert len(captured_warnings) == 0


def test_save():
    """
    Test the IDF.save() function using a filehandle to avoid external effects.
    """
    file_text = "Material,TestMaterial,  !- Name"
    idf = IDF(StringIO(file_text))
    # test save with just a filehandle
    file_handle = StringIO()
    idf.save(file_handle)
    expected = "TestMaterial"
    file_handle.seek(0)
    result = file_handle.read()
    # minimal test that TestMaterial is being written to the file handle
    assert expected in result


# Structure of a test:
#
# 1. Arrange -> pytest fixture
# 2. Act
# 3. Assert
# 4. Cleanup -> pytest fixture


@pytest.fixture
def createidfinafolder():
    """make a temp dir and save an idf in it

    :Teardown: remove the temp dir and it'a contents"""
    # make a temp dir
    tdirname = "./atempdir1"
    abspath = os.path.abspath(tdirname)
    os.mkdir(tdirname)
    # make blank file
    idftxt = ""  # empty string
    fhandle = StringIO(idftxt)
    idf = IDF(fhandle)
    # put in some objects - building and version
    building = idf.newidfobject("building")
    building.Name = "Taj"
    idf.newidfobject("version")
    # save it in subdir1
    fname = f"{tdirname}/a.idf"
    idf.saveas(fname)

    yield idf
    # Teardown
    # remove the temp dir
    shutil.rmtree(abspath)
    # abspath ensure removal from any dir


@pytest.fixture
def changedir():
    """make a temp dir and change to that dir

    :Teardown: return to original dir and delete this temp dir"""
    # make a temp dir
    origdir = os.path.abspath(os.path.curdir)
    tdirname = "./atempdir2"
    abspath = os.path.abspath(tdirname)
    os.mkdir(tdirname)
    # change to that directory
    os.chdir(tdirname)
    yield tdirname

    # Teardown
    # remove the temp dir
    os.chdir(origdir)
    shutil.rmtree(abspath)
    # abspath ensure removal from any dir


def test_save_with_dir_change(createidfinafolder, changedir):
    """py.test of save when dir has been changed"""
    change, expected = "Mahal", "Mahal"
    # createidfinafolder creates the idf
    idf = createidfinafolder
    idfabs = idf.idfabsname
    # changedir chnages dir
    newdir = changedir
    # make a change to the idf
    building = idf.idfobjects["building"][0]
    building.Name = change
    # and save while in new dir
    idf.save()  # should save in orig dir
    # read the file again
    idf1 = IDF(idfabs)
    building1 = idf1.idfobjects["building"][0]
    assert building1.Name == expected
    # test if it works with filepath.Path
    fname_path = Path(idfabs)
    assert isinstance(fname_path, Path)
    idf2 = IDF(fname_path)
    building2 = idf2.idfobjects["building"][0]
    assert building2.Name == expected


def test_save_with_lineendings_and_encodings():
    """
    Test the IDF.save() function with combinations of encodings and line
    endings.

    """
    file_text = "Material,TestMaterial,  !- Name"
    idf = IDF(StringIO(file_text))
    lineendings = ("windows", "unix", "default")
    encodings = ("ascii", "latin-1", "UTF-8")

    for le, enc in product(lineendings, encodings):
        file_handle = StringIO()
        idf.save(file_handle, encoding=enc, lineendings=le)
        file_handle.seek(0)
        result = file_handle.read().encode(enc)
        if le == "windows":
            assert b"\r\n" in result
        elif le == "unix":
            assert b"\r\n" not in result
        elif le == "default":
            assert os.linesep.encode(enc) in result


def test_saveas():
    """Test the IDF.saveas() function."""
    file_text = "Material,TestMaterial,  !- Name"
    idf = IDF(StringIO(file_text))
    idf.idfname = "test.idf"

    try:
        idf.saveas()  # this should raise an error as no filename is passed
        assert False
    except TypeError:
        pass

    file_handle = StringIO()
    idf.saveas(file_handle)  # save with a filehandle
    expected = "TestMaterial"
    file_handle.seek(0)
    result = file_handle.read()
    assert expected in result

    # test the idfname attribute has been changed
    assert idf.idfname != "test.idf"


def test_savecopy():
    """Test the IDF.savecopy() function."""
    file_text = "Material,TestMaterial,  !- Name"
    idf = IDF(StringIO(file_text))
    idf.idfname = "test.idf"

    try:
        idf.savecopy()  # this should raise an error as no filename is passed
        assert False
    except TypeError:
        pass

    file_handle = StringIO()
    idf.savecopy(file_handle)  # save a copy with a different filename
    expected = "TestMaterial"
    file_handle.seek(0)
    result = file_handle.read()
    assert expected in result

    # test the idfname attribute has not been changed
    assert idf.idfname == "test.idf"


def test_initread():
    """Test for IDF.initread() with filename in unicode and as python str."""
    # setup
    idf = IDF()
    idf.initreadtxt(idfsnippet)
    idf.saveas("tmp.idf")

    # test fname as unicode
    fname = "tmp.idf"
    assert isinstance(fname, string_types)
    idf = IDF()
    idf.initread(fname)
    assert idf.getobject("BUILDING", "Building")

    # test fname as str
    fname = str("tmp.idf")
    assert isinstance(fname, string_types)
    idf = IDF()
    idf.initread(fname)
    assert idf.getobject("BUILDING", "Building")

    # test that a nonexistent file raises an IOError
    fname = "notarealfilename.notreal"
    idf = IDF()
    try:
        idf.initread(fname)
        assert False  # shouldn't reach here
    except IOError:
        pass

    # teardown
    os.remove("tmp.idf")


def test_initreadtxt():
    """Test for IDF.initreadtxt()."""
    idftxt = """
        Material,
          G01a 19mm gypsum board,  !- Name
          MediumSmooth,            !- Roughness
          0.019,                   !- Thickness {m}
          0.16,                    !- Conductivity {W/m-K}
          800,                     !- Density {kg/m3}
          1090;                    !- Specific Heat {J/kg-K}

        Construction,
          Interior Wall,           !- Name
          G01a 19mm gypsum board,  !- Outside Layer
          F04 Wall air space resistance,  !- Layer 2
          G01a 19mm gypsum board;  !- Layer 3
        """
    idf = IDF()
    idf.initreadtxt(idftxt)
    assert idf.getobject("MATERIAL", "G01a 19mm gypsum board")


def test_idfstr():
    """Test all outputtype options in IDF.idfstr()."""
    idf = IDF()
    idf.initreadtxt(idfsnippet)
    assert idf.outputtype == "standard"  # start with the default
    original = idf.idfstr()
    assert "!-" in original  # has comment
    assert "\n" in original  # has line break
    assert "\n\n" in original  # has empty line

    idf.outputtype = "standard"
    s = idf.idfstr()
    assert "!-" in s  # has comment
    assert "\n" in s  # has line break
    assert "\n\n" in s  # has empty line
    assert s == original  # is unchanged

    idf.outputtype = "nocomment"
    s = idf.idfstr()
    assert "!-" not in s  # has no comments
    assert "\n" in s  # has line break
    assert "\n\n" in s  # has empty line
    assert s != original  # is changed

    idf.outputtype = "nocomment1"
    s = idf.idfstr()
    assert "!-" not in s  # has no comments
    assert "\n" in s  # has line break
    assert "\n\n" in s  # has empty lines
    assert s != original  # is changed

    idf.outputtype = "nocomment2"
    s = idf.idfstr()
    assert "!-" not in s  # has no comments
    assert "\n" in s  # has line break
    assert "\n\n" not in s  # has no empty lines
    assert s != original  # is changed

    idf.outputtype = "compressed"
    s = idf.idfstr()
    assert "!-" not in s  # has no comments
    assert "\n" not in s  # has no line breaks
    assert "\n\n" not in s  # has no empty lines
    assert s != original  # is changed


def test_refname2key():
    """py.test for refname2key"""
    tdata = (
        (
            "TransformerNames",
            ["ElectricLoadCenter:Distribution".upper()],
        ),  # refname, key
        (
            "AllCurves",
            [
                "PUMP:VARIABLESPEED",
                "PUMP:CONSTANTSPEED",
                "BOILER:HOTWATER",
                "ENERGYMANAGEMENTSYSTEM:CURVEORTABLEINDEXVARIABLE",
            ],
        ),  # refname, key
    )
    for refname, key in tdata:
        fhandle = StringIO("")
        idf = IDF(fhandle)
        result = modeleditor.refname2key(idf, refname)
        assert result == key


def test_getiddgroupdict():
    """py.test for IDF.getiddgroupdict()"""
    data = (({None: ["Lead Input", "Simulation Data"]},),)  # gdict,
    for (gdict,) in data:
        fhandle = StringIO("")
        idf = IDF(fhandle)
        result = idf.getiddgroupdict()
        assert result[None] == gdict[None]


def test_idfinmsequence():
    """py.test for setting of theidf in Idf_MSequence"""
    idftxt = """Version, 6.0;"""
    # theidf set in Idf_MSequence.__init__
    idf = IDF(StringIO(idftxt))
    versions = idf.idfobjects["version".upper()]
    assert versions.theidf == idf
    assert versions[0].theidf == idf
    # theidf set in Idf_MSequence.insert()
    material = idf.newidfobject("material".upper())
    assert material.theidf == idf
    # theidf set when you pop an item
    newmaterial = idf.newidfobject("material".upper())
    materials = idf.idfobjects["material".upper()]
    material = materials.pop(0)
    assert material.theidf == None
    assert materials[0].theidf == idf


def test_idd_index():
    """py.test to see if idd_index is returned"""
    idftxt = """"""
    idf = IDF(StringIO(idftxt))
    assert idf.idd_index == {}
